import json
from pathlib import Path
from logging import getLogger

from typing import Dict, Any, Union, Optional, List, Iterator
from typing_extensions import Literal
from functools import lru_cache

from .. import load_shared_data, get_shared_data_root

from .pipette_definition import (
    PipetteConfigurations,
    PipetteLiquidPropertiesDefinition,
    ValidNozzleMaps,
)
from .model_constants import MOUNT_CONFIG_LOOKUP_TABLE, _MAP_KEY_TO_V2
from .types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteGenerationType,
    PipetteVersionType,
    PipetteModelMajorVersion,
    PipetteModelMinorVersion,
    LiquidClasses,
    PipetteOEMType,
)


LoadedConfiguration = Dict[str, Union[str, Dict[str, Any]]]

LOG = getLogger(__name__)


def _get_configuration_dictionary(
    config_type: Literal["general", "geometry", "liquid"],
    channels: PipetteChannelType,
    model: PipetteModelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
    liquid_class: Optional[LiquidClasses] = None,
) -> LoadedConfiguration:
    oem_extension = f"_{oem.value}" if oem != PipetteOEMType.OT else ""
    if liquid_class:
        config_path = (
            get_shared_data_root()
            / "pipette"
            / "definitions"
            / "2"
            / config_type
            / f"{channels.name.lower()}{oem_extension}"
            / model.value
            / liquid_class.name
            / f"{version.major}_{version.minor}.json"
        )
    else:
        config_path = (
            get_shared_data_root()
            / "pipette"
            / "definitions"
            / "2"
            / config_type
            / f"{channels.name.lower()}{oem_extension}"
            / model.value
            / f"{version.major}_{version.minor}.json"
        )
    return json.loads(load_shared_data(config_path))


@lru_cache(maxsize=10)
def _geometry(
    channels: PipetteChannelType,
    model: PipetteModelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> LoadedConfiguration:
    return _get_configuration_dictionary("geometry", channels, model, version, oem)


@lru_cache(maxsize=10)
def _liquid(
    channels: PipetteChannelType,
    model: PipetteModelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> Dict[str, LoadedConfiguration]:
    liquid_dict = {}
    for liquid_class in LiquidClasses:
        try:
            liquid_dict[liquid_class.name] = _get_configuration_dictionary(
                "liquid", channels, model, version, oem, liquid_class
            )
        except FileNotFoundError:
            continue

    return liquid_dict


@lru_cache(maxsize=10)
def _physical(
    channels: PipetteChannelType,
    model: PipetteModelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> LoadedConfiguration:
    return _get_configuration_dictionary("general", channels, model, version, oem)


def _dirs_in(path: Path) -> Iterator[Path]:
    for child in path.iterdir():
        if child.is_dir():
            yield child


@lru_cache(maxsize=1)
def load_serial_lookup_table() -> Dict[str, str]:
    """Load a serial abbreviation lookup table mapped to model name."""
    config_path = get_shared_data_root() / "pipette" / "definitions" / "2" / "general"
    _lookup_table = {}
    _channel_shorthand = {
        "eight_channel": "M",
        "single_channel": "S",
        "ninety_six_channel": "H",
        "eight_channel_em": "P",
    }
    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }
    _model_shorthand = {"p1000": "p1k", "p300": "p3h", "p200": "p2h"}
    for channel_dir in _dirs_in(config_path):
        for model_dir in _dirs_in(channel_dir):
            for version_file in model_dir.iterdir():
                if version_file.suffix != ".json":
                    continue
                try:
                    version_list = version_file.stem.split("_")
                    built_model = f"{model_dir.stem}_{_channel_model_str[channel_dir.stem]}_v{version_list[0]}.{version_list[1]}"
                except IndexError:
                    LOG.warning(f"Pipette def with bad name {version_file} ignored")
                    continue
                model_shorthand = _model_shorthand.get(model_dir.stem, model_dir.stem)
                if (
                    model_dir.stem == "p300"
                    and int(version_list[0]) == 1
                    and int(version_list[1]) == 0
                ):
                    # Well apparently, we decided to switch the shorthand of the p300 depending
                    # on whether it's a "V1" model or not...so...here is the lovely workaround.
                    model_shorthand = model_dir.stem
                serial_shorthand = f"{model_shorthand.upper()}{_channel_shorthand[channel_dir.stem]}V{version_list[0]}{version_list[1]}"
                _lookup_table[serial_shorthand] = built_model
    return _lookup_table


def load_liquid_model(
    model: PipetteModelType,
    channels: PipetteChannelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> Dict[str, PipetteLiquidPropertiesDefinition]:
    liquid_dict = _liquid(channels, model, version, oem)
    return {
        k: PipetteLiquidPropertiesDefinition.model_validate(v)
        for k, v in liquid_dict.items()
    }


def _change_to_camel_case(c: str) -> str:
    # Tiny helper function to convert to camelCase.
    config_name = c.split("_")
    if len(config_name) == 1:
        return config_name[0]
    return f"{config_name[0]}" + "".join(s.capitalize() for s in config_name[1::])


def _edit_non_quirk_with_lc_override(
    mutable_config_key: str,
    new_mutable_value: Any,
    base_dict: Dict[str, Any],
    liquid_class: Optional[LiquidClasses],
) -> None:
    def _do_edit_non_quirk(
        new_value: Any, existing: Dict[Any, Any], keypath: List[Any]
    ) -> None:
        thiskey: Any = keypath[0]
        if thiskey in [lc.name for lc in LiquidClasses]:
            if liquid_class:
                thiskey = liquid_class
            else:
                thiskey = LiquidClasses[thiskey]
        if len(keypath) > 1:
            restkeys = keypath[1:]
            if thiskey == "##EACHTIP##":
                for key in existing.keys():
                    _do_edit_non_quirk(new_value, existing[key], restkeys)
            else:
                _do_edit_non_quirk(new_value, existing[thiskey], restkeys)
        else:
            # This was the last key
            if thiskey == "##EACHTIP##":
                for key in existing.keys():
                    existing[key] = new_value
            else:
                existing[thiskey] = new_value

    new_names = _MAP_KEY_TO_V2[mutable_config_key]
    _do_edit_non_quirk(new_mutable_value, base_dict, new_names)


def update_pipette_configuration(
    base_configurations: PipetteConfigurations,
    v1_configuration_changes: Dict[str, Any],
    liquid_class: Optional[LiquidClasses] = None,
) -> PipetteConfigurations:
    """Helper function to update 'V1' format configurations (left over from PipetteDict).

    #TODO (lc 7-14-2023) Remove once the pipette config dict is eliminated.
    Given an input of v1 mutable configs, look up the equivalent keyed
    value of that configuration."""
    quirks_list = []
    dict_of_base_model = base_configurations.model_dump(by_alias=True)

    for c, v in v1_configuration_changes.items():
        lookup_key = _change_to_camel_case(c)
        if c == "quirks" and isinstance(v, dict):
            quirks_list.extend([b.name for b in v.values() if b.value])
        else:
            _edit_non_quirk_with_lc_override(
                lookup_key, v, dict_of_base_model, liquid_class
            )

    dict_of_base_model["quirks"] = list(
        set(dict_of_base_model["quirks"]) - set(quirks_list)
    )

    # re-serialization is not great for this nested enum so we need
    # to perform this workaround.
    if not liquid_class:
        liquid_class = LiquidClasses.default
    dict_of_base_model["liquid_properties"][liquid_class]["supportedTips"] = {
        k.name: v
        for k, v in dict_of_base_model["liquid_properties"][liquid_class][
            "supportedTips"
        ].items()
    }
    dict_of_base_model["liquid_properties"] = {
        k.name: v for k, v in dict_of_base_model["liquid_properties"].items()
    }
    dict_of_base_model["plungerPositionsConfigurations"] = {
        k.name: v
        for k, v in dict_of_base_model["plungerPositionsConfigurations"].items()
    }
    return PipetteConfigurations.model_validate(dict_of_base_model)


def load_definition(
    model: PipetteModelType,
    channels: PipetteChannelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> PipetteConfigurations:
    if (
        version.major not in PipetteModelMajorVersion
        or version.minor not in PipetteModelMinorVersion
    ):
        raise KeyError("Pipette version not found.")

    geometry_dict = _geometry(channels, model, version, oem)
    physical_dict = _physical(channels, model, version, oem)
    liquid_dict = _liquid(channels, model, version, oem)

    generation = PipetteGenerationType(physical_dict["displayCategory"])
    mount_configs = MOUNT_CONFIG_LOOKUP_TABLE[generation][channels]

    return PipetteConfigurations.model_validate(
        {
            **geometry_dict,
            **physical_dict,
            "liquid_properties": liquid_dict,
            "version": version,
            "mount_configurations": mount_configs,
        }
    )


def load_valid_nozzle_maps(
    model: PipetteModelType,
    channels: PipetteChannelType,
    version: PipetteVersionType,
    oem: PipetteOEMType,
) -> ValidNozzleMaps:
    if (
        version.major not in PipetteModelMajorVersion
        or version.minor not in PipetteModelMinorVersion
    ):
        raise KeyError("Pipette version not found.")

    physical_dict = _physical(channels, model, version, oem)
    return ValidNozzleMaps.model_validate(physical_dict["validNozzleMaps"])



"""Tests that we have all errors in our python bindings."""

import json

import pytest

from opentrons_shared_data.load import load_shared_data
from opentrons_shared_data.errors.codes import ErrorCodes


error_json = json.loads(load_shared_data("errors/definitions/1/errors.json"))


@pytest.mark.parametrize("code", list(error_json["codes"].keys()))
def check_error_present_and_correct(code: str) -> None:
    """Test that each error in json is retrievable."""
    error = ErrorCodes.by_error_code(code)
    assert error.value.code == code
    assert error.value.category.value.name == error_json["code"]["category"]



import pytest
import typeguard


from opentrons_shared_data.robot import load
from opentrons_shared_data.robot.types import RobotDefinition, RobotType


@pytest.mark.parametrize("defname", ["OT-2 Standard", "OT-3 Standard"])
def test_v1_defs(defname: RobotType) -> None:
    defn = load(robot_type=defname, version=1)
    typeguard.check_type(defn, RobotDefinition)
    assert defn["robotType"] == defname



"""opentrons_shared_data.command: functions command schemas."""
from pathlib import Path
from functools import cache
import json
import os
import re
from logging import getLogger

from opentrons_shared_data.errors.exceptions import (
    InvalidProtocolData,
    PythonException,
    InvalidStoredData,
)

from ..load import load_shared_data, get_shared_data_root

LOG = getLogger(__name__)

SCHEMA_REF_VERSION_RE = re.compile(r"opentronsCommandSchemaV(\d+)")


def get_newest_schema_version() -> str:
    """Get the version string of the most modern command schema currently in shared-data."""
    schema_ids = known_schema_ids()
    max_version = 0
    for schema_id in schema_ids:
        ref_version_match = SCHEMA_REF_VERSION_RE.match(schema_id)
        if not ref_version_match:
            continue
        ref_version = int(ref_version_match.group(1))
        if ref_version > max_version:
            max_version = ref_version

    return str(max_version)


def load_schema_string(version: str) -> str:
    """Get the string containing the command JSON schema for the given version string."""
    path = Path("command") / "schemas" / f"{version}.json"
    try:
        return json.dumps(json.loads(load_shared_data(path)), indent=2)
    except OSError as ose:
        raise InvalidProtocolData(
            message=f"Command schema version {version} is not available",
            detail={
                "type": "bad-schema-version",
                "schema-kind": "command",
                "version": version,
            },
            wrapping=[PythonException(ose)],
        )


def schema_version_from_ref(ref: str) -> str:
    """Parse the command schema version from a command schema ref."""
    version = SCHEMA_REF_VERSION_RE.match(ref)
    if not version:
        raise InvalidProtocolData(
            message=f"Could not parse version from command schema ${ref}",
            detail={"ref": ref, "type": "bad-schema-ref", "schema-kind": "command"},
        )
    return version.group(1)


def _schema_id_from_file(schema_file_name: str) -> str:
    schema_version_match = re.match(r"^.*(\d+).json$", schema_file_name)
    if schema_version_match is None:
        raise InvalidStoredData(
            message=f"Command schema {schema_file_name} is not named appropriately for its version",
            detail={"type": "bad-schema-name", "schema-kind": "command"},
        )
    try:
        schema_content = json.load(open(schema_file_name))
    except json.JSONDecodeError as jde:
        raise InvalidStoredData(
            message=f"Command schema {schema_file_name} is not valid json",
            detail={"type": "bad-schema-json", "schema-kind": "command"},
            wrapping=[PythonException(jde)],
        ) from jde

    try:
        schema_id: str = schema_content["$id"]
    except KeyError as ke:
        raise InvalidStoredData(
            message=f"Command schema {schema_file_name} has no $id",
            detail={"type": "bad-schema-json", "schema-kind": "command"},
            wrapping=[PythonException(ke)],
        ) from ke
    if not SCHEMA_REF_VERSION_RE.match(schema_id):
        raise InvalidStoredData(
            message=f"Command schema {schema_file_name} has an invalid id {schema_id} that does not match opentronsCommandSchema#"
        )
    return schema_id


@cache
def known_schema_ids() -> list[str]:
    """Get a list of all known command schema IDs."""
    command_schemas_dir = get_shared_data_root() / "command" / "schemas"
    command_schemas = os.listdir(command_schemas_dir)
    all_schema_ids = []
    for schema_file_name in command_schemas:
        try:
            all_schema_ids.append(
                _schema_id_from_file(str(command_schemas_dir / schema_file_name))
            )
        except Exception:
            LOG.exception(
                f"Could not load command schema from {str(command_schemas_dir/schema_file_name)}, skipping"
            )
    return all_schema_ids



""" opentrons_shared_data.labware.types: types for labware defs

types in this file by and large require the use of typing_extensions.
this module shouldn't be imported unless typing.TYPE_CHECKING is true.
"""
from typing import NewType
from typing_extensions import Literal, TypedDict, NotRequired
from .constants import (
    CircularType,
    RectangularType,
)

LabwareUri = NewType("LabwareUri", str)

LabwareDisplayCategory = Literal[
    "tipRack",
    "tubeRack",
    "reservoir",
    "trash",
    "wellPlate",
    "aluminumBlock",
    "adapter",
    "other",
    "lid",
    "system",
]

LabwareFormat = Literal[
    "96Standard",
    "384Standard",
    "trough",
    "irregular",
    "trash",
]

LabwareRoles = Literal[
    "labware",
    "fixture",
    "adapter",
    "maintenance",
    "lid",
    "system",
]

SpringDirectionalForce = Literal["backLeftBottom"]


class Vector2D(TypedDict):
    x: float
    y: float


class Vector3D(TypedDict):
    x: float
    y: float
    z: float


class AxisAlignedBoundingBox2D(TypedDict):
    backLeft: Vector2D
    frontRight: Vector2D


class AxisAlignedBoundingBox3D(TypedDict):
    backLeftBottom: Vector3D
    frontRightTop: Vector3D


class GripperOffsets(TypedDict):
    pickUpOffset: Vector3D
    dropOffset: Vector3D


class LabwareParameters2(TypedDict):
    format: LabwareFormat
    isTiprack: bool
    loadName: str
    isMagneticModuleCompatible: bool
    isDeckSlotCompatible: NotRequired[bool]
    quirks: NotRequired[list[str]]
    tipLength: NotRequired[float]
    tipOverlap: NotRequired[float]
    magneticModuleEngageHeight: NotRequired[float]


class LabwareParameters3(LabwareParameters2, TypedDict):
    pass  # Currently equivalent to LabwareParameters2.


class LabwareBrandData(TypedDict):
    brand: str
    brandId: NotRequired[list[str]]
    links: NotRequired[list[str]]


class LabwareMetadata(TypedDict):
    displayName: str
    displayCategory: LabwareDisplayCategory
    displayVolumeUnits: Literal["µL", "mL", "L"]
    tags: NotRequired[list[str]]


class LabwareDimensions(TypedDict):
    yDimension: float
    zDimension: float
    xDimension: float


class _WellCommon2(TypedDict):
    depth: float
    totalLiquidVolume: float
    x: float
    y: float
    z: float
    geometryDefinitionId: NotRequired[str | None]


class CircularWellDefinition2(_WellCommon2, TypedDict):
    shape: CircularType
    diameter: float


class RectangularWellDefinition2(_WellCommon2, TypedDict):
    shape: RectangularType
    xDimension: float
    yDimension: float


WellDefinition2 = CircularWellDefinition2 | RectangularWellDefinition2


class CircularWellDefinition3(CircularWellDefinition2, TypedDict):
    # Currently equivalent to CircularWellDefinition2.
    pass


class RectangularWellDefinition3(RectangularWellDefinition2, TypedDict):
    # Currently equivalent to RectangularWellDefinition2.
    pass


WellDefinition3 = CircularWellDefinition3 | RectangularWellDefinition3


class WellGroupMetadata(TypedDict):
    displayName: NotRequired[str]
    displayCategory: NotRequired[LabwareDisplayCategory]
    wellBottomShape: NotRequired[Literal["flat", "u", "v"]]


class WellGroup(TypedDict):
    wells: list[str]
    metadata: WellGroupMetadata
    brand: NotRequired[LabwareBrandData]


class Extents(TypedDict):
    total: AxisAlignedBoundingBox3D


class SlotFootprintAsChildFeature(TypedDict):
    z: float
    backLeft: Vector2D
    frontRight: Vector2D
    springDirectionalForce: NotRequired[SpringDirectionalForce]


class SlotFootprintAsParentFeature(TypedDict):
    z: float
    backLeft: Vector2D
    frontRight: Vector2D
    springDirectionalForce: NotRequired[SpringDirectionalForce]


class OpentronsFlexTipRackLidAsParentFeature(TypedDict):
    matingZ: float


class OpentronsFlexTipRackLidAsChildFeature(TypedDict):
    matingZ: float


class LocatingFeatures(TypedDict):
    """A dictionary of locating features."""

    slotFootprintAsChild: NotRequired[SlotFootprintAsChildFeature]
    slotFootprintAsParent: NotRequired[SlotFootprintAsParentFeature]
    springDirectionalForceAsParent: NotRequired[SpringDirectionalForce]
    opentronsFlexTipRackLidAsParent: NotRequired[OpentronsFlexTipRackLidAsParentFeature]
    opentronsFlexTipRackLidAsChild: NotRequired[OpentronsFlexTipRackLidAsChildFeature]


class LabwareDefinition2(TypedDict):
    schemaVersion: Literal[2]
    version: int
    namespace: str
    metadata: LabwareMetadata
    brand: LabwareBrandData
    parameters: LabwareParameters2
    cornerOffsetFromSlot: Vector3D
    ordering: list[list[str]]
    dimensions: LabwareDimensions
    wells: dict[str, WellDefinition2]
    groups: list[WellGroup]
    stackingOffsetWithLabware: NotRequired[dict[str, Vector3D]]
    stackingOffsetWithModule: NotRequired[dict[str, Vector3D]]
    allowedRoles: NotRequired[list[LabwareRoles]]
    gripperOffsets: NotRequired[dict[str, GripperOffsets]]
    gripForce: NotRequired[float]
    gripHeightFromLabwareBottom: NotRequired[float]
    stackLimit: NotRequired[int]
    compatibleParentLabware: NotRequired[list[str]]
    # The innerLabwareGeometry dict values are not currently modeled in these
    # TypedDict-based bindings. The only code that cares about them
    # currentlyuses our Pydantic-based bindings instead.
    innerLabwareGeometry: NotRequired[dict[str, object] | None]


# Class to mix in the "$otSharedSchema" key. This cannot be defined with the normal
# TypedDict class syntax because it contains a dollar sign.
_OTSharedSchemaMixin = TypedDict(
    "_OTSharedSchemaMixin", {"$otSharedSchema": Literal["#/labware/schemas/3"]}
)


class LabwareDefinition3(_OTSharedSchemaMixin, TypedDict):
    schemaVersion: Literal[3]
    # $otSharedSchema mixed in via subclassing
    version: int
    namespace: str
    metadata: LabwareMetadata
    brand: LabwareBrandData
    parameters: LabwareParameters3
    ordering: list[list[str]]
    features: LocatingFeatures
    extents: Extents
    wells: dict[str, WellDefinition3]
    groups: list[WellGroup]
    stackingOffsetWithLabware: NotRequired[dict[str, Vector3D]]
    stackingOffsetWithModule: NotRequired[dict[str, Vector3D]]
    allowedRoles: NotRequired[list[LabwareRoles]]
    gripperOffsets: NotRequired[dict[str, GripperOffsets]]
    gripForce: NotRequired[float]
    gripHeightFromLabwareOrigin: NotRequired[float]
    stackLimit: NotRequired[int]
    compatibleParentLabware: NotRequired[list[str]]
    # The innerLabwareGeometry dict values are not currently modeled in these
    # TypedDict-based bindings. The only code that cares about them
    # currentlyuses our Pydantic-based bindings instead.
    innerLabwareGeometry: NotRequired[dict[str, object] | None]


LabwareDefinition = LabwareDefinition2 | LabwareDefinition3



from pydantic import (
    ConfigDict,
    BaseModel,
    Field,
    model_validator,
)
from typing import Any, List, Optional, Dict, Union
from typing_extensions import Literal
from opentrons_shared_data.labware.labware_definition import LabwareDefinition2

from .shared_models import (
    Liquid,
    Labware,
    CommandAnnotation,
    Location,
    ProfileStep,
    WellLocation,
    OffsetVector,
    Metadata,
    Module,
    Pipette,
    Robot,
    DesignerApplication,
)


# TODO (tamar 3/15/22): split apart all the command payloads when we tackle #9583
class Params(BaseModel):
    slotName: Optional[str] = None
    axes: Optional[List[str]] = None
    pipetteId: Optional[str] = None
    mount: Optional[str] = None
    moduleId: Optional[str] = None
    location: Optional[Union[Location, Literal["offDeck"]]] = None
    labwareId: Optional[str] = None
    displayName: Optional[str] = None
    liquidId: Optional[str] = None
    volumeByWell: Optional[Dict[str, Any]] = None
    wellName: Optional[str] = None
    volume: Optional[float] = None
    flowRate: Optional[float] = None
    wellLocation: Optional[Union[WellLocation]] = None
    waitForResume: Optional[Literal[True]] = None
    seconds: Optional[float] = None
    minimumZHeight: Optional[float] = None
    forceDirect: Optional[bool] = None
    speed: Optional[float] = None
    message: Optional[str] = None
    coordinates: Optional[OffsetVector] = None
    axis: Optional[str] = None
    distance: Optional[float] = None
    positionId: Optional[str] = None
    temperature: Optional[float] = None
    celsius: Optional[float] = None
    blockMaxVolumeUl: Optional[float] = None
    rpm: Optional[float] = None
    height: Optional[float] = None
    offset: Optional[OffsetVector] = None
    profile: Optional[List[ProfileStep]] = None
    radius: Optional[float] = None


class Command(BaseModel):
    commandType: str
    params: Params
    key: Optional[str] = None


class ProtocolSchemaV6(BaseModel):
    otSharedSchema: Literal["#/protocol/schemas/6"] = Field(
        ...,
        alias="$otSharedSchema",
        description="The path to a valid Opentrons shared schema relative to "
        "the shared-data directory, without its extension.",
    )
    schemaVersion: Literal[6]
    metadata: Metadata
    robot: Robot
    pipettes: Dict[str, Pipette]
    labware: Dict[str, Labware]
    modules: Optional[Dict[str, Module]] = None
    liquids: Optional[Dict[str, Liquid]] = None
    labwareDefinitions: Dict[str, LabwareDefinition2]
    # commands must be after pipettes, labware, etc. for its @validator to work.
    commands: List[Command]
    commandAnnotations: Optional[List[CommandAnnotation]] = None
    designerApplication: Optional[DesignerApplication] = None
    model_config = ConfigDict(populate_by_name=True)

    @model_validator(mode="after")
    def _validate_commands(self) -> "ProtocolSchemaV6":
        pipette_ids = set(self.pipettes.keys() if self.pipettes else [])
        labware_ids = set(self.labware.keys() if self.labware else [])
        module_ids = set(self.modules.keys() if self.modules else [])
        liquid_ids = set(self.liquids.keys() if self.liquids else [])

        for index, command in enumerate(self.commands):
            if (
                command.params.pipetteId is not None
                and command.params.pipetteId not in pipette_ids
            ):
                raise ValueError(
                    f"{command.commandType} command at index {index}"
                    f" references ID {command.params.pipetteId},"
                    f" which doesn't exist."
                )
            if (
                command.params.labwareId is not None
                and command.params.labwareId not in labware_ids
            ):
                raise ValueError(
                    f"{command.commandType} command at index {index}"
                    f" references ID {command.params.labwareId},"
                    f" which doesn't exist."
                )
            if (
                command.params.moduleId is not None
                and command.params.moduleId not in module_ids
            ):
                raise ValueError(
                    f"{command.commandType} command at index {index}"
                    f" references ID {command.params.moduleId},"
                    f" which doesn't exist."
                )
            if (
                command.params.liquidId is not None
                and command.params.liquidId not in liquid_ids
            ):
                raise ValueError(
                    f"{command.commandType} command at index {index}"
                    f" references ID {command.params.liquidId},"
                    f" which doesn't exist."
                )

        return self



"""
opentrons_shared_data.module.types: types requiring typing_extensions
for modules
"""

from typing import Any, Dict, List, Union
from typing_extensions import Literal, TypedDict

from opentrons_shared_data.labware.types import LocatingFeatures, Extents

SchemaV1 = Literal["1"]
SchemaV2 = Literal["2"]
SchemaV3 = Literal["3"]
SchemaVersions = Union[SchemaV1, SchemaV2, SchemaV3]

ModuleSchema = Dict[str, Any]

MagneticModuleType = Literal["magneticModuleType"]
TemperatureModuleType = Literal["temperatureModuleType"]
ThermocyclerModuleType = Literal["thermocyclerModuleType"]
HeaterShakerModuleType = Literal["heaterShakerModuleType"]
MagneticBlockType = Literal["magneticBlockType"]
AbsorbanceReaderType = Literal["absorbanceReaderType"]
FlexStackerModuleType = Literal["flexStackerModuleType"]

ModuleType = Union[
    MagneticModuleType,
    TemperatureModuleType,
    ThermocyclerModuleType,
    HeaterShakerModuleType,
    MagneticBlockType,
    AbsorbanceReaderType,
    FlexStackerModuleType,
]

MagneticModuleModel = Literal["magneticModuleV1", "magneticModuleV2"]
TemperatureModuleModel = Literal["temperatureModuleV1", "temperatureModuleV2"]
ThermocyclerModuleModel = Literal["thermocyclerModuleV1", "thermocyclerModuleV2"]
HeaterShakerModuleModel = Literal["heaterShakerModuleV1"]
MagneticBlockModel = Literal["magneticBlockV1"]
AbsorbanceReaderModel = Literal["absorbanceReaderV1"]
FlexStackerModuleModel = Literal["flexStackerModuleV1"]

ModuleModel = Union[
    MagneticModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
    HeaterShakerModuleModel,
    MagneticBlockModel,
    AbsorbanceReaderModel,
    FlexStackerModuleModel,
]

ModuleSlotTransform = TypedDict(
    "ModuleSlotTransform", {"labwareOffset": List[List[float]]}
)

ModuleLabwareOffset = TypedDict(
    "ModuleLabwareOffset", {"x": float, "y": float, "z": float}
)

ModuleDimensions = TypedDict(
    "ModuleDimensions",
    {
        "bareOverallHeight": float,
        "overLabwareHeight": float,
        "lidHeight": float,
        "xDimension": float,
        "yDimension": float,
        "footprintXDimension": float,
        "footprintYDimension": float,
        "labwareInterfaceXDimension": float,
        "labwareInterfaceYDimension": float,
        "maxStackerFillHeight": float,
        "maxStackerRetrievableHeight": float,
    },
    total=False,
)

ModuleCalibrationPointOffset = TypedDict(
    "ModuleCalibrationPointOffset", {"x": float, "y": float}
)

ModuleCalibrationPointOffsetWithZ = TypedDict(
    "ModuleCalibrationPointOffsetWithZ", {"x": float, "y": float, "z": float}
)

CornerOffsetFromSlot = TypedDict(
    "CornerOffsetFromSlot", {"x": float, "y": float, "z": float}
)


class NamedOffset(TypedDict):
    x: float
    y: float
    z: float


class GripperOffsets(TypedDict):
    pickUpOffset: NamedOffset
    dropOffset: NamedOffset


class TOFBaseline(TypedDict):
    extend: Dict[int, List[float]]
    retract: Dict[int, List[float]]


class TOFSensorBaseline(TypedDict):
    version: str
    X: TOFBaseline
    Z: TOFBaseline


# TODO(mc, 2022-03-18): potentially move from typed-dict to Pydantic
ModuleDefinitionV3 = TypedDict(
    "ModuleDefinitionV3",
    {
        "$otSharedSchema": Literal["module/schemas/3"],
        "moduleType": ModuleType,
        "model": ModuleModel,
        "labwareOffset": ModuleLabwareOffset,
        "cornerOffsetFromSlot": CornerOffsetFromSlot,
        "features": LocatingFeatures,
        "extents": Extents,
        "dimensions": ModuleDimensions,
        "calibrationPoint": ModuleCalibrationPointOffsetWithZ,
        "config": Dict[str, int],
        "displayName": str,
        "quirks": List[str],
        "slotTransforms": Dict[str, Dict[str, Dict[str, List[List[float]]]]],
        "compatibleWith": List[ModuleModel],
        "uniqueModuleData": Dict[str, Any],
        "incompatibleWithDecks": List[str],
        "gripperOffsets": Dict[str, GripperOffsets],
    },
    total=False,
)

# V2 is not used anymore. This type is preserved for historical purposes
ModuleDefinitionV2 = TypedDict(
    "ModuleDefinitionV2",
    {
        "$otSharedSchema": Literal["module/schemas/2"],
        "moduleType": ModuleType,
        "model": ModuleModel,
        "labwareOffset": ModuleLabwareOffset,
        "dimensions": ModuleDimensions,
        "calibrationPoint": ModuleCalibrationPointOffset,
        "displayName": str,
        "quirks": List[str],
        "slotTransforms": Dict[str, Dict[str, Dict[str, List[List[float]]]]],
        "compatibleWith": List[ModuleModel],
    },
)

ModuleDefinitionV1 = TypedDict(
    "ModuleDefinitionV1",
    {
        "labwareOffset": ModuleLabwareOffset,
        "dimensions": ModuleDimensions,
        "calibrationPoint": ModuleCalibrationPointOffset,
        "displayName": str,
        "loadName": str,
        "quirks": List[str],
    },
)



"""
opentrons_shared_data.pipette.types: types for pipette config that
require typing_extensions.

This module should only be imported if typing.TYPE_CHECKING is True.
"""
from enum import Enum
from typing import Dict, List, NewType, Union

from typing_extensions import Literal, TypedDict

# TODO(mc, 2022-06-16): remove type alias when able
# and when certain removal will not break any pickling
from ..labware.types import LabwareUri as LabwareUri


PipetteName = Literal[
    "p10_single",
    "p10_multi",
    "p20_single_gen2",
    "p20_multi_gen2",
    "p50_single",
    "p50_multi",
    "p50_single_flex",
    "p50_multi_flex",
    "p300_single",
    "p300_multi",
    "p300_single_gen2",
    "p300_multi_gen2",
    "p1000_single",
    "p1000_single_gen2",
    "p1000_single_flex",
    "p1000_multi_flex",
    "p1000_multi_em",
    "p1000_96",
    "p200_96",
]


class PipetteNameType(str, Enum):
    """Pipette load name values."""

    value: PipetteName

    P10_SINGLE = "p10_single"
    P10_MULTI = "p10_multi"
    P20_SINGLE_GEN2 = "p20_single_gen2"
    P20_MULTI_GEN2 = "p20_multi_gen2"
    P50_SINGLE = "p50_single"
    P50_MULTI = "p50_multi"
    P50_SINGLE_FLEX = "p50_single_flex"
    P50_MULTI_FLEX = "p50_multi_flex"
    P300_SINGLE = "p300_single"
    P300_MULTI = "p300_multi"
    P300_SINGLE_GEN2 = "p300_single_gen2"
    P300_MULTI_GEN2 = "p300_multi_gen2"
    P1000_SINGLE = "p1000_single"
    P1000_SINGLE_GEN2 = "p1000_single_gen2"
    P1000_SINGLE_FLEX = "p1000_single_flex"
    P1000_MULTI_FLEX = "p1000_multi_flex"
    P1000_MULTI_EM = "p1000_multi_em"
    P1000_96 = "p1000_96"
    P200_96 = "p200_96"


# Generic NewType for models because we get new ones frequently and theres
# a huge number of them
PipetteModel = NewType("PipetteModel", str)

DisplayCategory = Literal["GEN1", "GEN2", "FLEX"]

# todo(mm, 2022-03-18):
# The JSON schema defines this as any string, not as an enum of string literals.
# Check if it's safe to simplify this to just str.
ConfigUnit = Literal[
    "mm",
    "amps",
    "mm/sec",
    "mm/s",  # todo(mm, 2022-03-18): Standardize specs to mm/sec or mm/s.
    "presses",
]

Quirk = NewType("Quirk", str)

ChannelCount = Literal[1, 8, 96]

UlPerMmAction = Literal["aspirate", "dispense", "blowout"]


class PipetteConfigElement(TypedDict):
    value: float
    min: float
    max: float


class PipetteConfigElementWithPerApiLevelValue(TypedDict):
    value: float
    min: float
    max: float
    valuesByApiLevel: Dict[str, float]


# TypedDicts can't be generic sadly
class PipetteCustomizableConfigElementFloat(TypedDict):
    value: float
    min: float
    max: float
    units: ConfigUnit
    type: Literal["float"]


class PipetteCustomizableConfigElementInt(TypedDict):
    value: int
    min: int
    max: int
    units: ConfigUnit
    type: Literal["int"]


PipetteCustomizableConfigElement = Union[
    PipetteCustomizableConfigElementFloat, PipetteCustomizableConfigElementInt
]

SmoothieConfigs = TypedDict(
    "SmoothieConfigs",
    {"stepsPerMM": float, "homePosition": float, "travelDistance": float},
)


class PipetteNameSpec(TypedDict):
    displayName: str
    displayCategory: DisplayCategory
    minVolume: Union[float, int]
    maxVolume: Union[float, int]
    channels: ChannelCount
    defaultAspirateFlowRate: PipetteConfigElementWithPerApiLevelValue
    defaultDispenseFlowRate: PipetteConfigElementWithPerApiLevelValue
    defaultBlowOutFlowRate: PipetteConfigElementWithPerApiLevelValue
    smoothieConfigs: SmoothieConfigs
    defaultTipracks: List[LabwareUri]


PipetteNameSpecs = Dict[PipetteName, PipetteNameSpec]

UlPerMm = Dict[UlPerMmAction, List[List[float]]]


class PipetteModelSpec(TypedDict, total=False):
    name: PipetteName
    top: PipetteCustomizableConfigElementFloat
    bottom: PipetteCustomizableConfigElementFloat
    blowout: PipetteCustomizableConfigElementFloat
    dropTip: PipetteCustomizableConfigElementFloat
    pickUpCurrent: PipetteCustomizableConfigElementFloat
    pickUpDistance: PipetteCustomizableConfigElementFloat
    pickUpIncrement: PipetteCustomizableConfigElementFloat
    pickUpPresses: PipetteCustomizableConfigElementInt
    pickUpSpeed: PipetteCustomizableConfigElementFloat
    plungerCurrent: PipetteCustomizableConfigElementFloat
    dropTipCurrent: PipetteCustomizableConfigElementFloat
    dropTipSpeed: PipetteCustomizableConfigElementFloat
    modelOffset: List[float]
    nozzleOffset: List[float]
    ulPerMm: List[UlPerMm]
    tipOverlap: Dict[str, float]
    tipLength: PipetteCustomizableConfigElementFloat
    quirks: List[Quirk]
    # these keys are not present in some pipette definitions
    backCompatNames: List[PipetteName]
    idleCurrent: float
    returnTipHeight: float


class PipetteFusedSpec(PipetteNameSpec, PipetteModelSpec, total=False):
    pass


class PipetteModelSpecs(TypedDict):
    config: Dict[PipetteModel, PipetteModelSpec]
    mutableConfigs: List[str]
    validQuirks: List[str]



"""A python package implementing the errors in errors.json.

A note on this package: it is not codegenned but written by hand, and has to be upkept
by hand. It comes with tests to make sure this happens.

This seems a little ridiculous and kind of is, but what it lets us do is really simplify
our static analysis and dev setup.
"""

from .categories import ErrorCategories, ErrorCategory
from .codes import ErrorCodes, ErrorCode
from .exceptions import (
    EnumeratedError,
    PythonException,
    GeneralError,
    RoboticsControlError,
    RoboticsInteractionError,
)

__all__ = [
    "ErrorCategory",
    "ErrorCategories",
    "ErrorCode",
    "ErrorCodes",
    "EnumeratedError",
    "PythonException",
    "GeneralError",
    "RoboticsControlError",
    "RoboticsInteractionError",
]



import pytest

from typing import Union
from opentrons_shared_data.pipette.types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteVersionType,
    PipetteGenerationType,
    PipetteOEMType,
)
from opentrons_shared_data.pipette.types import PipetteModel, PipetteName
from opentrons_shared_data.pipette import (
    pipette_definition as pc,
    pipette_load_name_conversions as ps,
)


@pytest.mark.parametrize(
    argnames=["model", "output"],
    argvalues=[
        [
            "p50_single_v2.0",
            pc.PipetteModelVersionType(
                PipetteModelType.p50,
                PipetteChannelType.SINGLE_CHANNEL,
                PipetteVersionType(2, 0),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_multi_v1.0",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.EIGHT_CHANNEL,
                PipetteVersionType(1, 0),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_96_v1.0",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.NINETY_SIX_CHANNEL,
                PipetteVersionType(1, 0),
                PipetteOEMType.OT,
            ),
        ],
    ],
)
def test_convert_pipette_model(
    model: PipetteModel, output: pc.PipetteModelVersionType
) -> None:
    assert output == ps.convert_pipette_model(model)


@pytest.mark.parametrize(
    argnames=["model", "version", "output"],
    argvalues=[
        [
            "p50_single",
            "2.0",
            pc.PipetteModelVersionType(
                PipetteModelType.p50,
                PipetteChannelType.SINGLE_CHANNEL,
                PipetteVersionType(2, 0),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_multi",
            "3.3",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.EIGHT_CHANNEL,
                PipetteVersionType(3, 3),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_96",
            "1.1",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.NINETY_SIX_CHANNEL,
                PipetteVersionType(1, 1),
                PipetteOEMType.OT,
            ),
        ],
    ],
)
def test_convert_pipette_model_provided_version(
    model: PipetteModel, version: str, output: pc.PipetteModelVersionType
) -> None:
    assert output == ps.convert_pipette_model(model, version)


@pytest.mark.parametrize(
    argnames=["name", "output"],
    argvalues=[
        [
            "p50_single_gen2",
            pc.PipetteModelVersionType(
                PipetteModelType.p50,
                PipetteChannelType.SINGLE_CHANNEL,
                PipetteVersionType(2, 0),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_multi_flex",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.EIGHT_CHANNEL,
                PipetteVersionType(3, 5),
                PipetteOEMType.OT,
            ),
        ],
        [
            "p1000_96",
            pc.PipetteModelVersionType(
                PipetteModelType.p1000,
                PipetteChannelType.NINETY_SIX_CHANNEL,
                PipetteVersionType(3, 6),
                PipetteOEMType.OT,
            ),
        ],
    ],
)
def test_convert_pipette_name(
    name: PipetteName, output: pc.PipetteModelVersionType
) -> None:
    assert output == ps.convert_pipette_name(name)


@pytest.mark.parametrize(
    argnames=["model_type", "channels", "generation", "output", "oem"],
    argvalues=[
        [
            PipetteModelType.p50,
            PipetteChannelType.SINGLE_CHANNEL,
            PipetteGenerationType.GEN2,
            "p50_single_gen2",
            PipetteOEMType.OT,
        ],
        [
            PipetteModelType.p1000,
            PipetteChannelType.EIGHT_CHANNEL,
            PipetteGenerationType.GEN2,
            "p1000_multi_gen2",
            PipetteOEMType.OT,
        ],
        [
            # 96 channel has a unique "name" right now
            PipetteModelType.p1000,
            PipetteChannelType.NINETY_SIX_CHANNEL,
            PipetteGenerationType.FLEX,
            "p1000_96",
            PipetteOEMType.OT,
        ],
    ],
)
def test_model_version_type_string_version(
    model_type: PipetteModelType,
    channels: PipetteChannelType,
    generation: PipetteGenerationType,
    output: PipetteName,
    oem: PipetteOEMType,
) -> None:
    data = pc.PipetteNameType(
        pipette_type=model_type,
        pipette_channels=channels,
        pipette_generation=generation,
        oem_type=oem,
    )
    assert output == str(data)


@pytest.mark.parametrize(
    argnames=["model_type", "channels", "version", "output", "oem"],
    argvalues=[
        [
            PipetteModelType.p50,
            PipetteChannelType.SINGLE_CHANNEL,
            PipetteVersionType(1, 0),
            "p50_single_v1",
            PipetteOEMType.OT,
        ],
        [
            PipetteModelType.p1000,
            PipetteChannelType.EIGHT_CHANNEL,
            PipetteVersionType(2, 1),
            "p1000_multi_v2.1",
            PipetteOEMType.OT,
        ],
        [
            PipetteModelType.p1000,
            PipetteChannelType.NINETY_SIX_CHANNEL,
            PipetteVersionType(3, 3),
            "p1000_96_v3.3",
            PipetteOEMType.OT,
        ],
    ],
)
def test_name_type_string_generation(
    model_type: PipetteModelType,
    channels: PipetteChannelType,
    version: PipetteVersionType,
    output: PipetteModel,
    oem: PipetteOEMType,
) -> None:
    data = pc.PipetteModelVersionType(
        pipette_type=model_type,
        pipette_channels=channels,
        pipette_version=version,
        oem_type=oem,
    )
    assert output == str(data)


@pytest.mark.parametrize(
    argnames=["model_or_name", "valid"],
    argvalues=[
        [
            "p50_single",
            True,
        ],
        [
            "p50_sing",
            False,
        ],
        [
            "p100_multi",
            False,
        ],
        [
            "p1000_multi_v3.3",
            True,
        ],
    ],
)
def test_supported_pipette(
    model_or_name: Union[PipetteName, PipetteModel, None], valid: bool
) -> None:
    assert ps.supported_pipette(model_or_name) == valid



"""Exception hierarchy for error codes."""

from typing import Dict, Any, Optional, List, Iterator, Union, Sequence, overload
from logging import getLogger
from traceback import format_exception_only, format_tb
import inspect
import sys

from .codes import ErrorCodes
from .categories import ErrorCategories


log = getLogger(__name__)


class EnumeratedError(Exception):
    """The root class of error-code-bearing exceptions."""

    def __init__(
        self,
        code: ErrorCodes,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence["EnumeratedError"]] = None,
    ) -> None:
        """Build an EnumeratedError."""
        self.code = code
        self.message = message or ""
        self.detail = detail or {}
        self.wrapping = wrapping or []

    def __repr__(self) -> str:
        """Get a representative string for the exception."""
        return f"<{self.__class__.__name__}: code=<{self.code.value.code} {self.code.name}> message={self.message} detail={str(self.detail)}"

    def __str__(self) -> str:
        """Get a human-readable string."""
        _node = self.detail.get("node")
        return f'Error {self.code.value.code} {self.code.name} ({self.__class__.__name__}){f": {self.message}" if self.message else ""}{f" ({_node})" if _node else ""}'

    def __eq__(self, other: object) -> bool:
        """Compare if two enumerated errors match."""
        if not isinstance(other, EnumeratedError):
            return NotImplemented
        return (
            self.code == other.code
            and self.message == other.message
            and self.detail == other.detail
            and self.wrapping == other.wrapping
        )


class CommunicationError(EnumeratedError):
    """An exception indicating an unknown communications error.

    This can either be used directly (by not specifying an error code, in which case
    it will get the base communication error code) or as an intermediate parent in which
    case you should specify the code with a super call.
    """

    def __init__(
        self,
        code: Optional[ErrorCodes] = None,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CommunicationError."""
        if code and code not in code.of_category(
            ErrorCategories.HARDWARE_COMMUNICATION_ERROR
        ):
            log.error(
                f"Error {code.name} is inappropriate for a CommunicationError exception"
            )
        super().__init__(
            code or ErrorCodes.COMMUNICATION_ERROR, message, detail, wrapping
        )


class RoboticsControlError(EnumeratedError):
    """An exception indicating an unknown robotics control error.

    This can either be used directly (by not specifying an error code, in which case
    it will get the base robotics control error code) or as an intermediate parent in which
    case you should specify the code with a super call.
    """

    def __init__(
        self,
        code: Optional[ErrorCodes] = None,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a RoboticsControlError."""
        if code and code not in code.of_category(
            ErrorCategories.ROBOTICS_CONTROL_ERROR
        ):
            log.error(
                f"Error {code.name} is inappropriate for a RoboticsControlError exception"
            )

        super().__init__(
            code or ErrorCodes.ROBOTICS_CONTROL_ERROR, message, detail, wrapping
        )


class RoboticsInteractionError(EnumeratedError):
    """An exception indicating an unknown robotics interacion error.

    This can either be used directly (by not specifying an error code, in which case
    it will get the base robotics interaction error code) or as an intermediate parent in which
    case you should specify the code with a super call.
    """

    def __init__(
        self,
        code: Optional[ErrorCodes] = None,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a RoboticsInteractionError."""
        if code and code not in code.of_category(
            ErrorCategories.ROBOTICS_INTERACTION_ERROR
        ):
            log.error(
                f"Error {code.name} is inappropriate for a RoboticsInteractionError exception"
            )

        super().__init__(
            code or ErrorCodes.ROBOTICS_INTERACTION_ERROR, message, detail, wrapping
        )


class GeneralError(EnumeratedError):
    """An exception indicating an unknown general error.

    This can either be used directly (by not specifying an error code, in which case
    it will get the base general error code) or as an intermediate parent in which
    case you should specify the code with a super call.
    """

    def __init__(
        self,
        code: Optional[ErrorCodes] = None,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[Union[EnumeratedError, BaseException]]] = None,
    ) -> None:
        """Build a GeneralError."""
        if code and code not in code.of_category(ErrorCategories.GENERAL_ERROR):
            log.error(
                f"Error {code.name} is inappropriate for a GeneralError exception"
            )

        def _wrapped_excs() -> Iterator[EnumeratedError]:
            if not wrapping:
                return
            for exc in wrapping:
                if isinstance(exc, EnumeratedError):
                    yield exc
                else:
                    yield PythonException(exc)

        super().__init__(
            code or ErrorCodes.GENERAL_ERROR, message, detail, list(_wrapped_excs())
        )


def _exc_harvest_predicate(v: Any) -> bool:
    if inspect.isroutine(v):
        return False
    if inspect.ismethoddescriptor(v):
        return False
    # on python 3.11 and up we can check if things are method wrappers, which basic builtin
    # dunders like __add__ are, but until then we can't and also don't know this is real
    if sys.version_info.minor >= 11 and inspect.ismethodwrapper(v):  # type: ignore[attr-defined]
        return False
    return True


class PythonException(GeneralError):
    """An exception wrapping a base exception but with a GeneralError code and storing details."""

    def __init__(self, exc: BaseException) -> None:
        """Build a PythonException."""

        def _descend_exc_ctx(exc: BaseException) -> List[PythonException]:
            descendants: List[PythonException] = []
            if exc.__context__:
                descendants.append(PythonException(exc.__context__))
            if exc.__cause__:
                descendants.append(PythonException(exc.__cause__))
            return descendants

        base_details = {
            k: str(v)
            for k, v in inspect.getmembers(exc, _exc_harvest_predicate)
            if not k.startswith("_")
        }
        try:
            tb = exc.__traceback__
        except AttributeError:
            tb = None

        if tb:
            base_details["traceback"] = "\n".join(format_tb(tb))
        base_details["class"] = type(exc).__name__

        super().__init__(
            message="\n".join(format_exception_only(type(exc), exc)).strip(),
            detail=base_details,
            wrapping=_descend_exc_ctx(exc),
        )


class RobotInUseError(CommunicationError):
    """An error indicating that an action cannot proceed because another is in progress."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CanbusCommunicationError."""
        super().__init__(ErrorCodes.ROBOT_IN_USE, message, detail, wrapping)


class CanbusCommunicationError(CommunicationError):
    """An error indicating a problem with canbus communication."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CanbusCommunicationError."""
        super().__init__(
            ErrorCodes.CANBUS_COMMUNICATION_ERROR, message, detail, wrapping
        )


class InternalUSBCommunicationError(CommunicationError):
    """An error indicating a problem with internal USB communication - e.g. with the rear panel."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InternalUSBCommunicationError."""
        super().__init__(
            ErrorCodes.INTERNAL_USB_COMMUNICATION_ERROR, message, detail, wrapping
        )


class ModuleCommunicationError(CommunicationError):
    """An error indicating a problem with module communication."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CanbusCommunicationError."""
        super().__init__(
            ErrorCodes.CANBUS_COMMUNICATION_ERROR, message, detail, wrapping
        )


class CommandTimedOutError(CommunicationError):
    """An error indicating that a command timed out."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CommandTimedOutError."""
        super().__init__(ErrorCodes.COMMAND_TIMED_OUT, message, detail, wrapping)


class FirmwareUpdateFailedError(CommunicationError):
    """An error indicating that a firmware update failed."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FirmwareUpdateFailedError."""
        super().__init__(ErrorCodes.FIRMWARE_UPDATE_FAILED, message, detail, wrapping)


class InternalMessageFormatError(CommunicationError):
    """An error indicating that an internal message was formatted incorrectly."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InternalMesasgeFormatError."""
        super().__init__(
            ErrorCodes.INTERNAL_MESSAGE_FORMAT_ERROR, message, detail, wrapping
        )


class CANBusConfigurationError(CommunicationError):
    """An error indicating a misconfiguration of the CANbus."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CANBus Configuration Error."""
        super().__init__(
            ErrorCodes.CANBUS_CONFIGURATION_ERROR, message, detail, wrapping
        )


class CANBusBusError(CommunicationError):
    """An error indicating a low-level bus error on the CANbus like an error frame."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CANBus Bus Error."""
        super().__init__(ErrorCodes.CANBUS_BUS_ERROR, message, detail, wrapping)


class MotionFailedError(RoboticsControlError):
    """An error indicating that a motion failed."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MotionFailedError."""
        super().__init__(ErrorCodes.MOTION_FAILED, message, detail, wrapping)


class HomingFailedError(RoboticsControlError):
    """An error indicating that a homing failed."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a HomingFailedError."""
        super().__init__(ErrorCodes.HOMING_FAILED, message, detail, wrapping)


class StallOrCollisionDetectedError(RoboticsControlError):
    """An error indicating that a stall or collision occurred."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a StallOrCollisionDetectedError."""
        super().__init__(
            ErrorCodes.STALL_OR_COLLISION_DETECTED, message, detail, wrapping
        )


class FlexStackerStallError(RoboticsControlError):
    """An error indicating that a stall or collision occurred in the flex stacker."""

    def __init__(
        self,
        serial: str,
        axis: str,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FlexStackerStallError."""
        self.serial = serial
        self.axis = axis
        super().__init__(
            ErrorCodes.STACKER_STALL_OR_COLLISION_DETECTED,
            message,
            detail,
            wrapping,
        )


class MotionPlanningFailureError(RoboticsControlError):
    """An error indicating that motion planning failed."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MotionPlanningFailureError."""
        super().__init__(ErrorCodes.MOTION_PLANNING_FAILURE, message, detail, wrapping)


class PositionEstimationInvalidError(RoboticsControlError):
    """An error indicating that a command failed because position estimation was invalid."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PositionEstimationFailedError."""
        super().__init__(
            ErrorCodes.POSITION_ESTIMATION_INVALID, message, detail, wrapping
        )


class MoveConditionNotMetError(RoboticsControlError):
    """An error indicating that a move completed without its condition being met."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MoveConditionNotMetError."""
        super().__init__(
            ErrorCodes.MOVE_CONDITION_NOT_MET,
            message or "Move completed without its complete condition being met",
            detail,
            wrapping,
        )


class CalibrationStructureNotFoundError(RoboticsControlError):
    """An error indicating that a calibration square was not able to be found."""

    def __init__(
        self,
        structure_height: float,
        lower_limit: float,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CalibrationStructureNotFoundError."""
        super().__init__(
            ErrorCodes.CALIBRATION_STRUCTURE_NOT_FOUND,
            f"Structure height at z={structure_height}mm beyond lower limit: {lower_limit}.",
            detail,
            wrapping,
        )


class FailedGripperPickupError(RoboticsControlError):
    """Raised when the gripper expects to be holding an object, but the jaw is closed farther than expected."""

    def __init__(
        self,
        message: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FailedGripperPickupError."""
        super().__init__(
            ErrorCodes.FAILED_GRIPPER_PICKUP_ERROR,
            message or "Expected to grip labware, but none found.",
            details,
            wrapping,
        )


class EdgeNotFoundError(RoboticsControlError):
    """An error indicating that a calibration square edge was not able to be found."""

    def __init__(
        self,
        edge_name: str,
        stride: float,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a EdgeNotFoundError."""
        super().__init__(
            ErrorCodes.EDGE_NOT_FOUND,
            f"Edge {edge_name} could not be verified at {stride} mm resolution.",
            detail,
            wrapping,
        )


class EarlyCapacitiveSenseTrigger(RoboticsControlError):
    """An error indicating that a capacitive probe triggered too early."""

    def __init__(
        self,
        found: float,
        nominal: float,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a EarlyCapacitiveSenseTrigger."""
        super().__init__(
            ErrorCodes.EARLY_CAPACITIVE_SENSE_TRIGGER,
            f"Calibration triggered early at z={found}mm, expected {nominal}",
            detail,
            wrapping,
        )


class InaccurateNonContactSweepError(RoboticsControlError):
    """An error indicating that a capacitive sweep was inaccurate."""

    def __init__(
        self,
        found: float,
        nominal: float,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a InaccurateNonContactSweepError."""
        msg = (
            f"Calibration detected a slot width of {found:.3f}mm, "
            f"which is too far from the design width of {nominal:.3f}mm"
        )
        super().__init__(
            ErrorCodes.INACCURATE_NON_CONTACT_SWEEP,
            msg,
            detail,
            wrapping,
        )


class MisalignedGantryError(RoboticsControlError):
    """An error indicating that the robot's gantry and deck are misaligned."""

    def __init__(
        self,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MisalignedGantryError."""
        msg = "This machine is misaligned and requires maintenance."
        if detail:
            msg += str(detail)
        super().__init__(
            ErrorCodes.MISALIGNED_GANTRY,
            msg,
            detail,
            wrapping,
        )


class UnmatchedTipPresenceStates(RoboticsControlError):
    """An error indicating that a tip presence check resulted in two differing responses."""

    def __init__(
        self,
        states: Dict[int, int],
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an UnmatchedTipPresenceStatesError."""
        format_tip_state = {0: "not detected", 1: "detected"}
        msg = (
            f"Received two differing tip presence statuses."
            f" Rear Sensor tips: {format_tip_state[states[0]]}."
            f" Front Sensor tips: {format_tip_state[states[1]]}."
        )
        if detail:
            msg += str(detail)
        super().__init__(
            ErrorCodes.UNMATCHED_TIP_PRESENCE_STATES,
            msg,
            detail,
            wrapping,
        )


class PositionUnknownError(RoboticsControlError):
    """An error indicating that the robot's position is unknown."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a PositionUnknownError."""
        super().__init__(ErrorCodes.POSITION_UNKNOWN, message, detail, wrapping)


class ExecutionCancelledError(RoboticsControlError):
    """An error indicating that the robot's execution manager has been cancelled."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ExecutionCancelledError."""
        super().__init__(ErrorCodes.EXECUTION_CANCELLED, message, detail, wrapping)


class MotorDriverError(RoboticsControlError):
    """An error indicating that a motor driver is in error state."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a MotorDriverError."""
        super().__init__(ErrorCodes.MOTOR_DRIVER_ERROR, message, detail, wrapping)


class PipetteLiquidNotFoundError(RoboticsControlError):
    """Error raised if liquid sensing move completes without detecting liquid."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Initialize PipetteLiquidNotFoundError."""
        super().__init__(
            ErrorCodes.PIPETTE_LIQUID_NOT_FOUND,
            message,
            detail,
            wrapping,
        )


class TipHitWellBottomError(RoboticsControlError):
    """Error raised if tip hits bottom of well while trying to detect liquid level."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Initialize TipHitWellBottomError."""
        super().__init__(
            ErrorCodes.TIP_HIT_WELL_BOTTOM,
            message,
            detail,
            wrapping,
        )


class LabwareDroppedError(RoboticsInteractionError):
    """An error indicating that the gripper dropped labware it was holding."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a LabwareDroppedError."""
        super().__init__(ErrorCodes.LABWARE_DROPPED, message, detail, wrapping)


class TipPickupFailedError(RoboticsInteractionError):
    """An error indicating that a pipette failed to pick up a tip."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a TipPickupFailedError."""
        super().__init__(ErrorCodes.TIP_PICKUP_FAILED, message, detail, wrapping)


class TipDropFailedError(RoboticsInteractionError):
    """An error indicating that a pipette failed to drop a tip."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a TipPickupFailedError."""
        super().__init__(ErrorCodes.TIP_DROP_FAILED, message, detail, wrapping)


class UnexpectedTipRemovalError(RoboticsInteractionError):
    """An error indicating that a pipette did not have a tip when it should (aka it fell off)."""

    def __init__(
        self,
        action: str,
        pipette_name: str,
        mount: str,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an UnexpectedTipRemovalError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["pipette_name"] = pipette_name
        checked_detail["mount"] = mount
        message = f"Cannot perform {action} without a tip attached."
        super().__init__(
            ErrorCodes.UNEXPECTED_TIP_REMOVAL, message, checked_detail, wrapping
        )


class UnexpectedTipAttachError(RoboticsInteractionError):
    """An error indicating that a pipette had a tip when it shouldn't."""

    def __init__(
        self,
        action: str,
        pipette_name: str,
        mount: str,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an UnexpectedTipAttachError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["pipette_name"] = pipette_name
        checked_detail["mount"] = mount
        message = f"Cannot perform {action} with a tip already attached."
        super().__init__(ErrorCodes.UNEXPECTED_TIP_ATTACH, message, detail, wrapping)


class HepaUVFailedError(RoboticsInteractionError):
    """An error indicating that the HEPA UV module has errored."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an HepaUVFailedError."""
        super().__init__(ErrorCodes.HEPA_UV_FAILED, message, detail, wrapping)


class FlexStackerShuttleMissingError(RoboticsInteractionError):
    """An error indicating the Flex Stacker shuttle cannot be detected."""

    def __init__(
        self,
        serial: str,
        expected_state: str,
        shuttle_state: str,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FlexStackerStallError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["serial"] = serial
        checked_detail["expected_state"] = expected_state
        checked_detail["shuttle_state"] = shuttle_state
        if message is not None:
            checked_message = message
        else:
            checked_message = (
                "Flex Stacker shuttle not detected in state "
                f"{expected_state}, found {shuttle_state}."
            )
        super().__init__(
            ErrorCodes.STACKER_SHUTTLE_MISSING,
            checked_message,
            checked_detail,
            wrapping,
        )


class FlexStackerShuttleLabwareError(RoboticsInteractionError):
    """An error occurred during Flex Stacker shuttle labware detection."""

    def __init__(
        self,
        serial: str,
        shuttle_state: str,
        labware_expected: bool,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FlexStackerStallError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["serial"] = serial
        checked_detail["shuttle_state"] = shuttle_state
        checked_detail["labware_expected"] = labware_expected
        if message is not None:
            checked_message = message
        else:
            checked_message = (
                f"Labware {'not' if labware_expected else ''} detected on shuttle"
            )
        super().__init__(
            ErrorCodes.STACKER_SHUTTLE_LABWARE_FAILED,
            checked_message,
            checked_detail,
            wrapping,
        )


class FlexStackerHopperLabwareError(RoboticsInteractionError):
    """An error occurred when detecting labware inside the Flex Stacker hopper."""

    def __init__(
        self,
        serial: str,
        labware_expected: bool,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FlexStackerStallError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["serial"] = serial
        checked_detail["labware_expected"] = labware_expected
        if message is not None:
            checked_message = message
        else:
            checked_message = f"Labware {'not' if labware_expected else ''} detected in Flex Stacker hopper"
        super().__init__(
            ErrorCodes.STACKER_HOPPER_LABWARE_FAILED,
            checked_message,
            checked_detail,
            wrapping,
        )


class FlexStackerShuttleNotEmptyError(RoboticsInteractionError):
    """An error occurred when the Flex Stacker Shuttle is not empty when it should be."""

    def __init__(
        self,
        serial: str,
        shuttle_state: str,
        labware_expected: bool,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FlexStackerShuttleNotEmptyError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["serial"] = serial
        checked_detail["shuttle_state"] = shuttle_state
        checked_detail["labware_expected"] = labware_expected
        if message is not None:
            checked_message = message
        else:
            checked_message = f"Flex Stacker {serial} shuttle is not empty."
        super().__init__(
            ErrorCodes.STACKER_SHUTTLE_OCCUPIED,
            checked_message,
            checked_detail,
            wrapping,
        )


class FirmwareUpdateRequiredError(RoboticsInteractionError):
    """An error indicating that a firmware update is required."""

    def __init__(
        self,
        action: str,
        subsystems_to_update: List[Any],
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a FirmwareUpdateRequiredError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["identifier"] = action
        checked_detail["subsystems_to_update"] = subsystems_to_update
        message = f"Cannot perform {action} until {subsystems_to_update} are updated."
        super().__init__(ErrorCodes.FIRMWARE_UPDATE_REQUIRED, message, detail, wrapping)


class PipetteOverpressureError(RoboticsInteractionError):
    """An error indicating that a pipette experienced an overpressure event, likely because of a clog."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an PipetteOverpressureError."""
        super().__init__(ErrorCodes.PIPETTE_OVERPRESSURE, message, detail, wrapping)


class EStopActivatedError(RoboticsInteractionError):
    """An error indicating that the E-stop was activated."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an EStopActivatedError."""
        super().__init__(ErrorCodes.E_STOP_ACTIVATED, message, detail, wrapping)


class EStopNotPresentError(RoboticsInteractionError):
    """An error indicating that the E-stop is not present."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an EStopNotPresentError."""
        super().__init__(ErrorCodes.E_STOP_NOT_PRESENT, message, detail, wrapping)


class PipetteNotPresentError(RoboticsInteractionError):
    """An error indicating that the specified pipette is not present."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an PipetteNotPresentError."""
        super().__init__(ErrorCodes.PIPETTE_NOT_PRESENT, message, detail, wrapping)


class GripperNotPresentError(RoboticsInteractionError):
    """An error indicating that the specified gripper is not present."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a GripperNotPresentError."""
        super().__init__(ErrorCodes.GRIPPER_NOT_PRESENT, message, detail, wrapping)


class InvalidActuator(RoboticsInteractionError):
    """An error indicating that a specified actuator is not valid."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidActuator."""
        super().__init__(ErrorCodes.INVALID_ACTUATOR, message, detail, wrapping)


class ModuleNotPresent(RoboticsInteractionError):
    """An error indicating that a specific module was not present."""

    def __init__(
        self,
        identifier: str,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a ModuleNotPresentError."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["identifier"] = identifier
        checked_message = message or f"Module {identifier} is not present"
        super().__init__(
            ErrorCodes.MODULE_NOT_PRESENT, checked_message, checked_detail, wrapping
        )


class InvalidInstrumentData(RoboticsInteractionError):
    """An error indicating that instrument data is invalid."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidInstrumentData."""
        super().__init__(ErrorCodes.INVALID_INSTRUMENT_DATA, message, detail, wrapping)


class InvalidLiquidClassName(RoboticsInteractionError):
    """An error indicating that a liquid class name does not exist on a pipette."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidLiquidClassName."""
        super().__init__(
            ErrorCodes.INVALID_LIQUID_CLASS_NAME, message, detail, wrapping
        )


class TipDetectorNotFound(RoboticsInteractionError):
    """An error indicating that a tip detector has not been created for a pipette."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a TipDetectorNotFound."""
        super().__init__(ErrorCodes.TIP_DETECTOR_NOT_FOUND, message, detail, wrapping)


class APIRemoved(GeneralError):
    """An error indicating that a specific API is no longer available."""

    @overload
    def __init__(  # noqa: D107
        self,
        *,
        api_element: Optional[str] = None,
        since_version: Optional[str] = None,
        current_version: Optional[str] = None,
        extra_message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        pass

    @overload
    def __init__(  # noqa: D107
        self,
        message: Optional[str] = None,
        *,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        pass

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        api_element: Optional[str] = None,
        since_version: Optional[str] = None,
        current_version: Optional[str] = None,
        extra_message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an APIRemoved error."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["identifier"] = api_element
        checked_detail["since_version"] = since_version
        checked_detail["current_version"] = current_version

        checked_api_element = api_element if api_element is not None else "This feature"

        if message is not None:
            checked_message = message
        else:
            if since_version is not None and current_version is not None:
                checked_message = (
                    f"{checked_api_element} is not available after API version {since_version}."
                    f" You are currently using API version {current_version}."
                )
            elif since_version is not None and current_version is None:
                checked_message = f"{checked_api_element} is not available after API version {since_version}."
            elif since_version is None and current_version is not None:
                checked_message = f"{checked_api_element} is not available in API version {current_version}."
            else:
                checked_message = f"{checked_api_element} is no longer available in the API version in use."

            if extra_message is not None:
                checked_message += " " + extra_message

        super().__init__(
            ErrorCodes.API_REMOVED, checked_message, checked_detail, wrapping
        )


class IncorrectAPIVersion(GeneralError):
    """An error indicating that a command was issued that is not supported by the API version in use."""

    @overload
    def __init__(  # noqa: D107
        self,
        *,
        api_element: Optional[str] = None,
        until_version: Optional[str] = None,
        current_version: Optional[str] = None,
        extra_message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        pass

    @overload
    def __init__(  # noqa: D107
        self,
        message: Optional[str] = None,
        *,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        pass

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        api_element: Optional[str] = None,
        until_version: Optional[str] = None,
        current_version: Optional[str] = None,
        extra_message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an IncorrectAPIVersion error."""
        checked_detail: Dict[str, Any] = detail or {}
        checked_detail["identifier"] = api_element
        checked_detail["until_version"] = until_version
        checked_detail["current_version"] = current_version

        checked_api_element = api_element if api_element is not None else "This feature"

        if message is not None:
            checked_message = message
        else:
            if until_version is not None and current_version is not None:
                checked_message = (
                    f"{checked_api_element} is not available until API version {until_version}."
                    f" You are currently using API version {current_version}."
                )
            elif until_version is not None and current_version is None:
                checked_message = f"{checked_api_element} is not available until API version {until_version}."
            elif until_version is None and current_version is not None:
                checked_message = f"{checked_api_element} is not available in API version {current_version}."
            else:
                checked_message = f"{checked_api_element} is not yet available in the API version in use."

            if extra_message is not None:
                checked_message += " " + extra_message

        super().__init__(
            ErrorCodes.INCORRECT_API_VERSION, checked_message, checked_detail, wrapping
        )


class CommandPreconditionViolated(GeneralError):
    """An error indicating that a command was issued in a robot state incompatible with it."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build a CommandPreconditionViolated instance."""
        super().__init__(
            ErrorCodes.COMMAND_PRECONDITION_VIOLATED, message, detail, wrapping
        )


class CommandParameterLimitViolated(GeneralError):
    """An error indicating that a command's parameter limit was violated."""

    def __init__(
        self,
        command_name: str,
        parameter_name: str,
        limit_statement: str,
        actual_value: str,
        wrapping: Optional[Sequence[Union[EnumeratedError, BaseException]]] = None,
    ) -> None:
        """Build a CommandParameterLimitViolated error."""
        wrapping_checked = wrapping or []
        super().__init__(
            ErrorCodes.COMMAND_PARAMETER_LIMIT_VIOLATED,
            f"The parameter {parameter_name} of {command_name} must be {limit_statement} but is {actual_value}",
            detail={
                "command": command_name,
                "argument": parameter_name,
                "limit": limit_statement,
                "value": actual_value,
            },
            wrapping=[
                PythonException(e) if isinstance(e, BaseException) else e
                for e in wrapping_checked
            ],
        )


class UnsupportedHardwareCommand(GeneralError):
    """An error indicating that a command being executed is not supported by the hardware."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an UnsupportedHardwareCommand."""
        super().__init__(
            ErrorCodes.NOT_SUPPORTED_ON_ROBOT_TYPE, message, detail, wrapping
        )


class InvalidProtocolData(GeneralError):
    """An error indicating that some protocol data is invalid."""

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[Union[EnumeratedError, BaseException]]] = None,
    ) -> None:
        """Build an InvalidProtocolData."""
        super().__init__(ErrorCodes.INVALID_PROTOCOL_DATA, message, detail, wrapping)


class InvalidStoredData(GeneralError):
    """An error indicating that some stored data is invalid.

    This will usually be because it was saved by a future version of the software.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an InvalidStoredData."""
        super().__init__(ErrorCodes.INVALID_STORED_DATA, message, detail, wrapping)


class MissingConfigurationData(GeneralError):
    """An error indicating that provided configuration data is missing or invalid.

    This will usually be because a pipette configuration does not match the ones provided by the pipette definition.
    """

    def __init__(
        self,
        message: Optional[str] = None,
        detail: Optional[Dict[str, str]] = None,
        wrapping: Optional[Sequence[EnumeratedError]] = None,
    ) -> None:
        """Build an MissingConfigurationData."""
        super().__init__(
            ErrorCodes.MISSING_CONFIGURATION_DATA, message, detail, wrapping
        )



from opentrons_shared_data.load import get_shared_data_root
from opentrons_shared_data.command import known_schema_ids


def test_known_schema_ids() -> None:
    schema_paths = get_shared_data_root() / "command" / "schemas"
    names = [filename.stem for filename in schema_paths.iterdir()]
    ids = [f"opentronsCommandSchemaV{version}" for version in names]
    assert sorted(ids) == sorted(known_schema_ids())



import json

from opentrons_shared_data import load_shared_data
from opentrons_shared_data.liquid_classes import load_definition, definition_exists
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
    PositionReference,
    Coordinate,
    TipPosition,
    Submerge,
    DelayParams,
    DelayProperties,
)


def test_load_liquid_class_schema_v1() -> None:
    fixture_data = load_shared_data("liquid-class/definitions/1/water/1.json")
    liquid_class_model = LiquidClassSchemaV1.model_validate_json(fixture_data)
    liquid_class_def_from_model = json.loads(
        liquid_class_model.model_dump_json(exclude_unset=True)
    )
    expected_liquid_class_def = json.loads(fixture_data)
    assert liquid_class_def_from_model == expected_liquid_class_def


def test_load_definition() -> None:
    water_definition = load_definition(name="water", version=1, schema_version=1)
    assert type(water_definition) is LiquidClassSchemaV1
    assert water_definition.byPipette[0].pipetteModel == "flex_1channel_50"
    assert water_definition.byPipette[0].byTipType[0].aspirate.submerge == Submerge(
        startPosition=TipPosition(
            positionReference=PositionReference.WELL_TOP,
            offset=Coordinate(x=0, y=0, z=2),
        ),
        speed=100,
        delay=DelayProperties(enable=False, params=DelayParams(duration=0)),
    )


def test_definition_exists() -> None:
    """Should return whether specified definition exists in shared data or not."""
    assert definition_exists(name="water", version=1) is True
    assert definition_exists(name="glycerol_50", version=1) is True
    assert definition_exists(name="glycerol_oh_no", version=1) is False
    assert definition_exists(name="glycerol_50", version=2) is False



import pytest
import typeguard

from typing import Generator, Tuple

from opentrons_shared_data.pipette import (
    model_config,
    name_config,
    fuse_specs,
    dummy_model_for_name,
)
from opentrons_shared_data.pipette.types import (
    PipetteModelSpecs,
    PipetteNameSpecs,
    PipetteFusedSpec,
    PipetteName,
    PipetteModel,
)


def test_model_config_check() -> None:
    defdict = model_config()
    typeguard.check_type(defdict, PipetteModelSpecs)


def test_name_config_check() -> None:
    defdict = name_config()
    typeguard.check_type(defdict, PipetteNameSpecs)


def build_model_name_pairs() -> Generator[Tuple[PipetteModel, PipetteName], None, None]:
    for model, conf in model_config()["config"].items():
        yield model, conf["name"]
        for bcn in conf.get("backCompatNames", []):
            yield model, bcn


@pytest.mark.parametrize("model,name", list(build_model_name_pairs()))
def test_fuse(model: PipetteModel, name: PipetteName) -> None:
    defdict = fuse_specs(model, name)
    typeguard.check_type(defdict, PipetteFusedSpec)


@pytest.mark.parametrize("name", list(name_config().keys()))
def test_model_for_name(name: PipetteName) -> None:
    model = dummy_model_for_name(name)
    assert model in model_config()["config"]



import re
from typing_extensions import Final
from typing import Literal, Union


# Regular expression to validate and extract row, column from well name
# (ie A3, C1)
WELL_NAME_PATTERN: Final["re.Pattern[str]"] = re.compile(r"^([A-Z]+)([0-9]+)$", re.X)

# These shapes are for wellshape definitions and describe the top of the well
CircularType = Literal["circular"]
Circular: CircularType = "circular"
RectangularType = Literal["rectangular"]
Rectangular: RectangularType = "rectangular"
WellShape = Union[Literal["circular"], Literal["rectangular"]]

# These shapes are used to describe the 3D primatives used to build wells
Conical = Literal["conical"]
Cuboidal = Literal["cuboidal"]
SquaredCone = Literal["squaredcone"]
RoundedCuboid = Literal["roundedcuboid"]
Spherical = Literal["spherical"]



import json
from opentrons_shared_data.gripper import load_definition, load_schema
from opentrons_shared_data.gripper.gripper_definition import (
    GripperModel,
    GripperDefinition,
)
from opentrons_shared_data import load_shared_data


def test_load_schema() -> None:
    assert load_schema(1) == json.loads(load_shared_data("gripper/schemas/1.json"))


def test_load_definition() -> None:
    gripper_def = load_definition(GripperModel.v1, 1)
    assert type(gripper_def) is GripperDefinition
    assert gripper_def.model == GripperModel.v1
    assert gripper_def.grip_force_profile.default_grip_force == 15
    assert gripper_def.geometry.base_offset_from_mount == (19.5, -74.325, -94.825)



"""A script for updating pre-existing V2 Pipette configurations."""

import os
import json
import argparse

from pathlib import Path
from typing import List, Dict, Tuple, Any, Iterator, Type

from pydantic import BaseModel
from pydantic.main import ModelMetaclass
from enum import Enum

from opentrons_shared_data import get_shared_data_root

from ..pipette_definition import (
    PipetteConfigurations,
    PipetteGeometryDefinition,
    PipettePhysicalPropertiesDefinition,
    PipetteLiquidPropertiesDefinition,
    PipetteModelVersionType,
    SupportedTipsDefinition,
)
from ..types import (
    PipetteModelType,
    PipetteChannelType,
    PipetteVersionType,
    PipetteTipType,
    PipetteModelMajorVersion,
    PipetteModelMinorVersion,
    LiquidClasses,
)
from ..load_data import _geometry, _physical, _liquid
from ..pipette_load_name_conversions import convert_pipette_model
from ..types import PipetteModel

"""
Instructions:

To run this script, you must be in `shared-data`. To invoke, use the command:

`pipenv run python -m opentrons_shared_data.pipette.scripts.update_configuration_files`

If you want to update all files, you can simply use the argument `--update_all_models`.

Make sure to run `make format-js` afterwards to ensure formatting of the json files
is good.

*Note* If you are adding a brand-new key, you MUST update the pydantic models
found in `python/pipette/pipette_definition.py` before running this script.

*Note* When you are entering in your data, please utilize the exact type. I.e. if it's a
list, you must input the list like: [1, 2, 3] or if it's a dict, like: {"data": 1}..

For now, we do not support updating pipetting functions in this script.
"""

ROOT = get_shared_data_root() / "pipette" / "definitions" / "2"
NOZZLE_LOCATION_CONFIGS = ["nozzle_offset", "nozzle_map"]


def _change_to_camel_case(c: str) -> str:
    # Tiny helper function to convert to camelCase.
    config_name = c.split("_")
    if len(config_name) == 1:
        return config_name[0]
    return f"{config_name[0]}" + "".join(s.capitalize() for s in config_name[1::])


def list_configuration_keys() -> Tuple[List[str], Dict[int, str]]:
    """List out the model keys available to modify at the top level."""
    lookup = {i: v for (i, v) in enumerate(PipetteConfigurations.model_fields)}
    return [
        f"{i}: {v}" for (i, v) in enumerate(PipetteConfigurations.model_fields)
    ], lookup


def list_available_enum(enum_type: Type[Enum]) -> List[str]:
    """List available pipette models"""
    return [f"{i}: {v}" for (i, v) in enumerate(enum_type)]


def handle_subclass_model(
    top_level_configuration: List[str], base_model: BaseModel, is_basemodel: bool
) -> List[str]:
    """Handle sub-classed basemodels and update the top level model as necessary."""
    if is_basemodel:
        if base_model.model_fields == SupportedTipsDefinition.model_fields:
            # pydantic does something weird with the types in ModelFields so
            # we cannot use isinstance checks to confirm if the base model
            # is a supported tips definition
            print(f"choose {PipetteTipType.__name__}:")
            for row in list_available_enum(PipetteTipType):
                print(f"\t{row}")
            tip_type = list(PipetteTipType)[
                int(input("select the tip volume size to modify"))
            ]
            top_level_configuration.append(tip_type.name)

        lookup = {i: v for (i, v) in enumerate(base_model.model_fields)}
        config_list = [f"{i}: {v}" for (i, v) in enumerate(base_model.model_fields)]
        print(f"you selected the basemodel {base_model.__name__}:")  # type: ignore[attr-defined]
        for row in config_list:
            print(f"\t{row}")

        configuration_to_update = lookup[
            int(input("select a specific configuration from above\n"))
        ]
        field_type = base_model.model_fields[
            configuration_to_update
        ].rebuild_annotation()
        is_basemodel = isinstance(field_type, ModelMetaclass)

        top_level_configuration.append(configuration_to_update)
        return handle_subclass_model(top_level_configuration, field_type, is_basemodel)
    else:
        return top_level_configuration


def check_from_version(version: str) -> str:
    """Check that the version requested is supported in the system."""
    version_int = [int(v) for v in version.split(".")]
    if version_int[0] not in PipetteModelMajorVersion:
        raise ValueError(f"Major version {version_int[0]} is not supported.")
    if version_int[1] not in PipetteModelMinorVersion:
        raise ValueError(f"Minor version {version_int[1]} is not supported.")
    return version


def save_data_to_file(
    directorypath: Path,
    file_name: str,
    data: Dict[str, Any],
) -> None:
    """
    Function used to save data to a file
    """
    directorypath.mkdir(parents=True, exist_ok=True)
    filepath = directorypath / f"{file_name}.json"
    with open(filepath, "w") as f:
        json.dump(data, f, indent=2)


def update(
    dict_to_update: Dict[str, Any], iter_of_configs: Iterator[str], value_to_update: Any
) -> Dict[str, Any]:
    """
    Recursively update the given dictionary to ensure no data is lost when updating.
    """
    next_key = next(iter_of_configs, None)
    if next_key and isinstance(dict_to_update.get(next_key), dict):
        dict_to_update[next_key] = update(
            dict_to_update.get(next_key, {}), iter_of_configs, value_to_update
        )
    elif next_key:
        dict_to_update[next_key] = value_to_update
    return dict_to_update


def build_nozzle_map(
    nozzle_offset: List[float], channels: PipetteChannelType
) -> Dict[str, List[float]]:
    Y_OFFSET = -9
    X_OFFSET = 9
    if channels == PipetteChannelType.SINGLE_CHANNEL:
        return {"A1": nozzle_offset}
    elif channels == PipetteChannelType.EIGHT_CHANNEL:
        return {
            f"{chr(ord('A') + 1*row)}1": [
                nozzle_offset[0],
                nozzle_offset[1] + Y_OFFSET * row,
                nozzle_offset[2],
            ]
            for row in range(8)
        }
    elif channels == PipetteChannelType.NINETY_SIX_CHANNEL:
        return {
            f"{chr(ord('A') + 1*row)}{1 + 1*col}": [
                nozzle_offset[0] + X_OFFSET * col,
                nozzle_offset[1] + Y_OFFSET * row,
                nozzle_offset[2],
            ]
            for row in range(8)
            for col in range(12)
        }
    raise ValueError(f"Unsupported channel type {channels}")


def load_and_update_file_from_config(
    config_to_update: List[str],
    value_to_update: Any,
    model_to_update: PipetteModelVersionType,
) -> None:
    """Update the requested config and save to disk.

    Load the requested config sub type (physical, geometry or liquid). Then
    update the current file and save to disk.

    """
    camel_list_to_update = iter([_change_to_camel_case(i) for i in config_to_update])

    if config_to_update[0] in PipetteGeometryDefinition.model_fields:
        geometry = _geometry(
            model_to_update.pipette_channels,
            model_to_update.pipette_type,
            model_to_update.pipette_version,
        )
        if config_to_update[0] == "nozzle_map":
            nozzle_to_use = (
                value_to_update if value_to_update else geometry["nozzleOffset"]
            )
            geometry["nozzleMap"] = build_nozzle_map(
                nozzle_to_use, model_to_update.pipette_channels
            )
        elif config_to_update[0] == "nozzle_offset":
            geometry["nozzleMap"] = build_nozzle_map(
                value_to_update, model_to_update.pipette_channels
            )
            geometry["nozzleOffset"] = value_to_update
        else:
            geometry = update(geometry, camel_list_to_update, value_to_update)
        PipetteGeometryDefinition.model_validate(geometry)

        filepath = (
            ROOT
            / "geometry"
            / model_to_update.pipette_channels.name.lower()
            / model_to_update.pipette_type.value
        )
        save_data_to_file(
            filepath,
            f"{model_to_update.pipette_version.major}_{model_to_update.pipette_version.minor}",
            geometry,
        )
    elif config_to_update[0] in PipettePhysicalPropertiesDefinition.model_fields:
        physical = _physical(
            model_to_update.pipette_channels,
            model_to_update.pipette_type,
            model_to_update.pipette_version,
        )

        physical = update(physical, camel_list_to_update, value_to_update)

        PipettePhysicalPropertiesDefinition.model_validate(physical)
        filepath = (
            ROOT
            / "general"
            / model_to_update.pipette_channels.name.lower()
            / model_to_update.pipette_type.value
        )
        save_data_to_file(
            filepath,
            f"{model_to_update.pipette_version.major}_{model_to_update.pipette_version.minor}",
            physical,
        )
    elif config_to_update[0] == "liquid_properties":
        next(camel_list_to_update)
        liquid = _liquid(
            model_to_update.pipette_channels,
            model_to_update.pipette_type,
            model_to_update.pipette_version,
        )

        print(
            "Please select what liquid class you wish to update.\n If you want to update all liquid classes then type 'all'.\n"
        )

        print(f"choose {LiquidClasses.__name__} or type 'all':")
        for row in list_available_enum(LiquidClasses):
            print(f"\t{row}")
        liquid_classes = input("liquid class: ")
        if liquid_classes == "all":
            for c in LiquidClasses:
                liquid = update(
                    liquid[c.name.lower()], camel_list_to_update, value_to_update
                )

                PipetteLiquidPropertiesDefinition.model_validate(liquid)
                filepath = (
                    ROOT
                    / "liquid"
                    / model_to_update.pipette_channels.name.lower()
                    / model_to_update.pipette_type.value
                    / c.name.lower()
                )
                save_data_to_file(
                    filepath,
                    f"{model_to_update.pipette_version.major}_{model_to_update.pipette_version.minor}",
                    liquid,
                )
        else:
            lc = list(LiquidClasses)[int(liquid_classes)]
            liquid = update(
                liquid[lc.name.lower()], camel_list_to_update, value_to_update
            )
            PipetteLiquidPropertiesDefinition.model_validate(liquid)

            filepath = (
                ROOT
                / "liquid"
                / model_to_update.pipette_channels.name.lower()
                / model_to_update.pipette_type.value
                / lc.name.lower()
            )
            save_data_to_file(
                filepath,
                f"{model_to_update.pipette_version.major}_{model_to_update.pipette_version.minor}",
                liquid,
            )
    else:
        raise KeyError(
            f"{config_to_update} is not saved to a file. Check `pipette_definition.py` for more information."
        )


def _update_single_model(configuration_to_update: List[str]) -> None:
    """Helper function to update single model."""
    print(f"choose {PipetteModelType.__name__}:")
    for row in list_available_enum(PipetteModelType):
        print(f"\t{row}")
    model = list(PipetteModelType)[int(input("Please select from above\n"))]

    print(f"choose {PipetteChannelType.__name__}:")
    for row in list_available_enum(PipetteChannelType):
        print(f"\t{row}")
    channels = list(PipetteChannelType)[int(input("Please select from above\n"))]

    version = PipetteVersionType.convert_from_float(
        float(check_from_version(input("Please input the version of the model\n")))
    )

    built_model: PipetteModel = PipetteModel(
        f"{model.name}_{str(channels)}_v{version.major}.{version.minor}"
    )

    if configuration_to_update[0] == NOZZLE_LOCATION_CONFIGS[1]:
        print(
            "You selected nozzle_map to edit. If you wish to update the nozzle offset, enter it on the next line.\n"
        )
        print("Otherwise, please type 'null' on the next line.\n")

    value_to_update = json.loads(
        input(
            f"Please select what you would like to update {configuration_to_update[-1]} to for {built_model}\n"
        )
    )

    model_version = convert_pipette_model(built_model)

    load_and_update_file_from_config(
        configuration_to_update, value_to_update, model_version
    )


def _update_all_models(configuration_to_update: List[str]) -> None:
    paths_to_validate = ROOT / "general"
    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }

    for channel_dir in os.listdir(paths_to_validate):
        for model_dir in os.listdir(paths_to_validate / channel_dir):
            for version_file in os.listdir(paths_to_validate / channel_dir / model_dir):
                version_list = version_file.split(".json")[0].split("_")
                built_model: PipetteModel = PipetteModel(
                    f"{model_dir}_{_channel_model_str[channel_dir]}_v{version_list[0]}.{version_list[1]}"
                )

                if configuration_to_update[0] == NOZZLE_LOCATION_CONFIGS[1]:
                    print(
                        "You selected nozzle_map to edit. If you wish to update the nozzle offset, enter it on the next line.\n"
                    )
                    print("Otherwise, please type 'null' on the next line.\n")

                value_to_update = json.loads(
                    input(
                        f"Please select what you would like to update {configuration_to_update} to for {built_model}\n"
                    )
                )

                model_version = convert_pipette_model(built_model)
                load_and_update_file_from_config(
                    configuration_to_update, value_to_update, model_version
                )


def determine_models_to_update(update_all_models: bool) -> None:
    try:
        while True:
            print(f"choose {PipetteConfigurations.__name__}:")
            config_list, table_lookup = list_configuration_keys()
            for row in config_list:
                print(f"\t{row}")

            configuration_to_update = [
                table_lookup[int(input("select a configuration from above\n"))]
            ]

            if configuration_to_update[0] == NOZZLE_LOCATION_CONFIGS[0]:
                print(
                    f"NOTE: updating the {configuration_to_update[0]} will automatically update the {NOZZLE_LOCATION_CONFIGS[1]}\n"
                )

            field_type = PipetteConfigurations.model_fields[
                configuration_to_update[0]
            ].rebuild_annotation()
            is_basemodel = isinstance(field_type, ModelMetaclass)

            configuration_to_update = handle_subclass_model(
                configuration_to_update, field_type, is_basemodel
            )
            if update_all_models:
                _update_all_models(configuration_to_update)
            else:
                _update_single_model(configuration_to_update)
    except KeyboardInterrupt:
        print("Finished updating! Validate that your files updated successfully.")


def main() -> None:
    """Entry point."""
    parser = argparse.ArgumentParser(
        description="96 channel tip handling testing script."
    )
    parser.add_argument(
        "--update_all_models",
        type=bool,
        help="update all",
        default=False,
    )

    args = parser.parse_args()

    determine_models_to_update(args.update_all_models)


if __name__ == "__main__":
    """
    A script to automate building a pipette configuration definition.

    This script can either perform migrations from a v1 -> v2 schema format
    or build a brand new script from scratch.

    When building a new pipette configuration model, you will either need
    to provide CSVs or use command line inputs.

    If you choose CSVs you will need one CSV for the general pipette configuration
    data (such as pipette model or number of channels) and one for every tip
    type that this pipette model can support.
    """
    main()



"""
opentrons_shared_data.labware: types and functions for accessing labware defs
"""

from __future__ import annotations

from functools import lru_cache
import json
from itertools import chain
from typing import Any, Iterable, NewType, TYPE_CHECKING, overload, Literal

from opentrons_shared_data.load import get_shared_data_root

from .. import load_shared_data

if TYPE_CHECKING:
    from .types import LabwareDefinition, LabwareDefinition2, LabwareDefinition3

Schema = NewType("Schema", dict[str, Any])


@overload
def load_definition(
    loadname: str, version: int, schema: Literal[2] = 2
) -> LabwareDefinition2:
    ...


@overload
def load_definition(
    loadname: str, version: int, schema: Literal[3]
) -> LabwareDefinition3:
    ...


def load_definition(loadname: str, version: int, schema: int = 2) -> LabwareDefinition:
    return json.loads(
        load_shared_data(f"labware/definitions/{schema}/{loadname}/{version}.json")
    )


def load_schema() -> Schema:
    return json.loads(load_shared_data("labware/schemas/2.json"))


@lru_cache(maxsize=1)
def list_definitions(schemas: Iterable[int] = [2, 3]) -> set[tuple[str, int, int]]:
    """Return (load_name, version, schema) for all labware definitions in shared-data."""

    def list_definitions_for_schema(schema: int) -> Iterable[tuple[str, int, int]]:
        root_for_schema = (
            get_shared_data_root() / "labware" / "definitions" / str(schema)
        )
        files = root_for_schema.glob("*/*.json")
        return ((file.parent.name, int(file.stem), schema) for file in files)

    return set(
        chain.from_iterable(list_definitions_for_schema(schema) for schema in schemas)
    )



from typing import TYPE_CHECKING
from enum import Enum


if TYPE_CHECKING:
    from .types import (
        DelayCommandId,
        BlowoutCommandId,
        PickUpTipCommandId,
        DropTipCommandId,
        AspirateCommandId,
        DispenseCommandId,
        TouchTipCommandId,
        MoveToSlotCommandId,
        AirGapCommandId,
        MagneticModuleEngageCommandId,
        MagneticModuleDisengageCommandId,
        TemperatureModuleSetTargetCommandId,
        TemperatureModuleAwaitCommandId,
        TemperatureModuleDeactivateCommandId,
        ThermocyclerSetTargetBlockCommandId,
        ThermocyclerSetTargetLidCommandId,
        ThermocyclerAwaitBlockTemperatureCommandId,
        ThermocyclerAwaitLidTemperatureCommandId,
        ThermocyclerOpenLidCommandId,
        ThermocyclerCloseLidCommandId,
        ThermocyclerDeactivateLidCommandId,
        ThermocyclerDeactivateBlockCommandId,
        ThermocyclerRunProfileCommandId,
        ThermocyclerAwaitProfileCommandId,
        MoveToWellCommandId,
    )


class JsonPipetteCommand(Enum):
    airGap: "AirGapCommandId" = "airGap"
    blowout: "BlowoutCommandId" = "blowout"
    pickUpTip: "PickUpTipCommandId" = "pickUpTip"
    dropTip: "DropTipCommandId" = "dropTip"
    aspirate: "AspirateCommandId" = "aspirate"
    dispense: "DispenseCommandId" = "dispense"
    touchTip: "TouchTipCommandId" = "touchTip"
    moveToSlot: "MoveToSlotCommandId" = "moveToSlot"
    moveToWell: "MoveToWellCommandId" = "moveToWell"


class JsonRobotCommand(Enum):
    delay: "DelayCommandId" = "delay"


class JsonMagneticModuleCommand(Enum):
    magneticModuleEngageMagnet: "MagneticModuleEngageCommandId" = (
        "magneticModule/engageMagnet"
    )
    magneticModuleDisengageMagnet: "MagneticModuleDisengageCommandId" = (
        "magneticModule/disengageMagnet"
    )


class JsonTemperatureModuleCommand(Enum):
    temperatureModuleSetTargetTemperature: "TemperatureModuleSetTargetCommandId" = (
        "temperatureModule/setTargetTemperature"
    )
    temperatureModuleDeactivate: "TemperatureModuleDeactivateCommandId" = (
        "temperatureModule/deactivate"
    )
    temperatureModuleAwaitTemperature: "TemperatureModuleAwaitCommandId" = (
        "temperatureModule/awaitTemperature"
    )


class JsonThermocyclerCommand(Enum):
    thermocyclerSetTargetBlockTemperature: "ThermocyclerSetTargetBlockCommandId" = (
        "thermocycler/setTargetBlockTemperature"
    )
    thermocyclerSetTargetLidTemperature: "ThermocyclerSetTargetLidCommandId" = (
        "thermocycler/setTargetLidTemperature"
    )
    thermocyclerAwaitBlockTemperature: "ThermocyclerAwaitBlockTemperatureCommandId" = (
        "thermocycler/awaitBlockTemperature"
    )
    thermocyclerAwaitLidTemperature: "ThermocyclerAwaitLidTemperatureCommandId" = (
        "thermocycler/awaitLidTemperature"
    )
    thermocyclerOpenLid: "ThermocyclerOpenLidCommandId" = "thermocycler/openLid"
    thermocyclerCloseLid: "ThermocyclerCloseLidCommandId" = "thermocycler/closeLid"
    thermocyclerDeactivateBlock: "ThermocyclerDeactivateBlockCommandId" = (
        "thermocycler/deactivateBlock"
    )
    thermocyclerDeactivateLid: "ThermocyclerDeactivateLidCommandId" = (
        "thermocycler/deactivateLid"
    )
    thermocyclerRunProfile: "ThermocyclerRunProfileCommandId" = (
        "thermocycler/runProfile"
    )
    thermocyclerAwaitProfileComplete: "ThermocyclerAwaitProfileCommandId" = (
        "thermocycler/awaitProfileComplete"
    )



import re
from functools import lru_cache
from typing import List, Optional, Union, cast, Literal, Tuple
from opentrons_shared_data import get_shared_data_root
from .types import PipetteModel, PipetteName

from .types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteVersionType,
    PipetteGenerationType,
    PipetteModelMajorVersionType,
    PipetteModelMinorVersionType,
    PipetteOEMType,
)
from .pipette_definition import (
    PipetteNameType,
    PipetteModelVersionType,
)

DEFAULT_CALIBRATION_OFFSET = [0.0, 0.0, 0.0]
DEFAULT_MODEL = PipetteModelType.p1000
DEFAULT_CHANNELS = PipetteChannelType.SINGLE_CHANNEL
DEFAULT_MODEL_VERSION = PipetteVersionType(major=1, minor=0)
DEFAULT_OEM = PipetteOEMType.OT

PIPETTE_AVAILABLE_TYPES = [m.name for m in PipetteModelType]
PIPETTE_CHANNELS_INTS = [c.value for c in PipetteChannelType]


def is_model(model_or_name: Union[PipetteName, PipetteModel, None]) -> bool:
    """Determine if we have a real model or just a PipetteName.

    Args:
        model_or_name (Union[PipetteName, PipetteModel, None]): The pipette we want to check.

    Returns:
        bool: Whether or not the given string is a PipetteModel
    """
    if not model_or_name:
        return False
    return "v" in model_or_name


def supported_pipette(model_or_name: Union[PipetteName, PipetteModel, None]) -> bool:
    """Determine if a pipette type is supported.

    Args:
        model_or_name (Union[PipetteName, PipetteModel, None]): The pipette we want to check.

    Returns:
        bool: Whether or not the given pipette name or model is supported.
    """
    if not model_or_name:
        return False
    split_model_or_name = model_or_name.split("_")
    try:
        channels_as_int = int(channels_from_string(split_model_or_name[1]))
    except ValueError:
        channels_as_int = 0
    if (
        split_model_or_name[0] in PIPETTE_AVAILABLE_TYPES
        and channels_as_int in PIPETTE_CHANNELS_INTS
    ):
        return True
    return False


def channels_from_string(channels: str) -> PipetteChannelType:
    """Convert channels from a string.

    With both `py:data:PipetteName` and `py:data:PipetteObject`, we refer to channel types
    as `single`, `multi` or `96`.

    Args:
        channels (str): The channel string we wish to convert.

    Returns:
        PipetteChannelType: A `py:obj:PipetteChannelType`
        representing the number of channels on a pipette.

    """
    if channels == "96":
        return PipetteChannelType.NINETY_SIX_CHANNEL
    elif "multi" in channels:
        return PipetteChannelType.EIGHT_CHANNEL
    elif channels == "single":
        return PipetteChannelType.SINGLE_CHANNEL
    else:
        raise ValueError("Invalid number of channels")


def version_from_string(version: str) -> PipetteVersionType:
    """Convert a version string to a py:obj:PipetteVersionType.

    The version string will either be in the format of `int.int` or `vint.int`.

    Args:
        version (str): The string version we wish to convert.

    Returns:
        PipetteVersionType: A pipette version object.

    """
    version_list = [v for v in re.split("\\.|[v]", version) if v]
    major = cast(PipetteModelMajorVersionType, int(version_list[0]))
    if len(version_list) > 1:
        minor = cast(PipetteModelMinorVersionType, int(version_list[1]))
    else:
        minor = 0
    return PipetteVersionType(major, minor)


def get_channel_from_pipette_name(pipette_name_tuple: Tuple[str, ...]) -> str:
    if "single" in pipette_name_tuple:
        return "single_channel"
    elif "96" in pipette_name_tuple:
        return "ninety_six_channel"
    else:
        if "em" in pipette_name_tuple:
            return "eight_channel_em"
        return "eight_channel"


def get_major_version_from_pipette_name(
    pipette_name_tuple: Tuple[str, ...],
) -> Literal[1, 2, 3]:
    #   special-casing for 96-channel to return version 3
    if (
        "flex" in pipette_name_tuple
        or "gen3" in pipette_name_tuple
        or "96" in pipette_name_tuple
    ):
        return 3
    elif "gen2" in pipette_name_tuple:
        return 2
    else:
        return 1


@lru_cache(4)
def version_from_generation(pipette_name_tuple: Tuple[str, ...]) -> PipetteVersionType:
    """Convert pipetteName to a py:obj:PipetteVersionType

    Given the pipette_name_tuple, cycle through each definition file path
    and find the latest version (major and minor version combined) that
    exists and return that version.

    Args:
        pipette_name_tuple (Tuple[str, ...]): A tuple of strings from the separated
        by `_` py:data:PipetteName.

    Returns:
        PipetteVersionType: A pipette version object.
    """
    major_version_from_pipette_name = get_major_version_from_pipette_name(
        pipette_name_tuple
    )
    model_from_pipette_name = pipette_name_tuple[0]
    channel_from_pipette_name = get_channel_from_pipette_name(pipette_name_tuple)
    paths_to_validate = (
        get_shared_data_root() / "pipette" / "definitions" / "2" / "general"
    )
    version_paths = (
        paths_to_validate / channel_from_pipette_name / model_from_pipette_name
    )

    highest_minor_version: PipetteModelMinorVersionType = 0

    for version_file in version_paths.iterdir():
        version_list = version_file.stem.split("_")
        major_version = version_list[0]
        minor_version = version_list[1]

        # Check if the major version matches the expected major version
        if major_version == str(major_version_from_pipette_name):
            minor_version_int = int(minor_version)
            minor_version_lit: PipetteModelMinorVersionType = cast(
                PipetteModelMinorVersionType, minor_version_int
            )

            # Update the highest minor version if this version is higher
            if highest_minor_version < minor_version_lit:
                highest_minor_version = minor_version_lit

    return PipetteVersionType(major_version_from_pipette_name, highest_minor_version)


def generation_from_string(pipette_name_list: List[str]) -> PipetteGenerationType:
    """Convert a string generation name to a py:obj:PipetteGenerationType.

    Args:
        pipette_name_list (List[str]): A list of strings from the separated by `_`
        py:data:PipetteName or py:data:PipetteModel.

    Returns:
        PipetteGenerationType: A pipette version object.

    """
    if (
        "flex" in pipette_name_list
        or "3." in pipette_name_list[-1]
        or pipette_name_list == ["p1000", "96"]
        or pipette_name_list == ["p200", "96"]
    ):
        return PipetteGenerationType.FLEX
    elif "gen2" in pipette_name_list or "2." in pipette_name_list[-1]:
        return PipetteGenerationType.GEN2
    else:
        return PipetteGenerationType.GEN1


def convert_to_pipette_name_type(
    model_or_name: Union[PipetteName, PipetteModel]
) -> PipetteNameType:
    """Convert the py:data:PipetteName to a py:obj:PipetteModelVersionType.

    `PipetteNames` are in the format of "p300_single" or "p300_single_gen1".

    Args:
        name (PipetteName): The pipette name we want to convert.

    Returns:
        PipetteNameType: An object representing a broken out PipetteName
        string.

    """
    split_pipette_model_or_name = model_or_name.split("_")
    channels = channels_from_string(split_pipette_model_or_name[1])
    generation = generation_from_string(split_pipette_model_or_name)
    pipette_type = PipetteModelType[split_pipette_model_or_name[0]]
    oem = PipetteOEMType.get_oem_from_model_str(model_or_name)
    return PipetteNameType(pipette_type, channels, generation, oem)


def convert_pipette_name(
    name: PipetteName, provided_version: Optional[str] = None
) -> PipetteModelVersionType:
    """Convert the py:data:PipetteName to a py:obj:PipetteModelVersionType.

    `PipetteNames` are in the format of "p300_single" or "p300_single_gen1".

    Args:
        name (PipetteName): The pipette name we want to convert.

    Returns:
        PipetteModelVersionType: An object representing a broken out PipetteName
        string.

    """
    split_pipette_name = name.split("_")
    channels_type = split_pipette_name[1]
    if len(split_pipette_name) > 2:
        if split_pipette_name[2] == "em":
            channels_type = "multi_em"

    channels = channels_from_string(channels_type)
    if provided_version:
        version = version_from_string(provided_version)
    else:
        pipette_name_tuple: Tuple[str, str, str] = (
            split_pipette_name[0],
            split_pipette_name[1],
            split_pipette_name[2] if len(split_pipette_name) > 2 else "",
        )
        version = version_from_generation(pipette_name_tuple)

    pipette_type = PipetteModelType[split_pipette_name[0]]
    oem = PipetteOEMType.get_oem_from_model_str(name)
    return PipetteModelVersionType(pipette_type, channels, version, oem)


def convert_pipette_model(
    model: Optional[PipetteModel], provided_version: Optional[str] = ""
) -> PipetteModelVersionType:
    """Convert the py:data:PipetteModel to a py:obj:PipetteModelVersionType.

    `PipetteModel` are in the format of "p300_single_v1.0" or "p300_single_v3.3".

    Sometimes, models may not have a version, in which case the `provided_version` arg
    allows you to specify a version to search for.

    Args:
        model (PipetteModel): The pipette model we want to convert.
        provided_version (str, Optional): The provided version we'd like to look for.

    Returns:
        PipetteModelVersionType: An object representing a broken out PipetteName
        string.

    """
    # TODO (lc 12-5-2022) This helper function is needed
    # until we stop using "name" and "model" to refer
    # to attached pipettes.
    # We need to figure out how to default the pipette model as well
    # rather than returning a p1000
    if model and not provided_version:
        # pipette_type, parsed_channels, parsed_version = model.split("_")
        exploded = model.split("_")
        if len(exploded) == 3:
            (pipette_type, parsed_channels, parsed_version) = exploded
            channels = channels_from_string(parsed_channels)
        else:
            pipette_type, parsed_channels, parsed_oem, parsed_version = exploded
            channels = channels_from_string(f"{parsed_channels}_{parsed_oem}")
        version = version_from_string(parsed_version)
        oem = PipetteOEMType.get_oem_from_model_str(str(model))
    elif model and provided_version:
        pipette_type, parsed_channels = model.split("_")
        channels = channels_from_string(parsed_channels)
        version = version_from_string(provided_version)
        oem = PipetteOEMType.get_oem_from_model_str(str(model))
    else:
        pipette_type = DEFAULT_MODEL.value
        channels = DEFAULT_CHANNELS
        version = DEFAULT_MODEL_VERSION
        oem = DEFAULT_OEM
    return PipetteModelVersionType(
        PipetteModelType[pipette_type], channels, version, oem
    )



"""opentrons_shared_data.robot.types: types for robot def."""
import enum
from typing import NewType, List, Dict, Any
from typing_extensions import Literal, TypedDict, NotRequired

RobotSchemaVersion1 = Literal[1]

RobotSchema = NewType("RobotSchema", Dict[str, Any])

RobotType = Literal["OT-2 Standard", "OT-3 Standard"]


class RobotTypeEnum(enum.Enum):
    """An enum representing the active robot type."""

    # TODO we should switch over to using Enums fully (if possible)
    # to represent our robot types in code -- rather than having
    # to string match everywhere.
    OT2 = enum.auto()
    FLEX = enum.auto()

    @classmethod
    def robot_literal_to_enum(cls, robot_type: RobotType) -> "RobotTypeEnum":
        """Convert Robot Type Literal to Robot Type Enum."""
        if robot_type == "OT-2 Standard":
            return cls.OT2
        elif robot_type == "OT-3 Standard":
            return cls.FLEX
        # No final `else` statement, depend on mypy exhaustiveness checking


class mountOffset(TypedDict):
    """The mount offsets for a given robot type based off the center of the carriage.."""

    left: List[float]
    right: List[float]
    gripper: NotRequired[List[float]]


class paddingOffset(TypedDict):
    """The padding offsets for a given robot type based off how far the pipettes can travel beyond the deck extents."""

    rear: float
    front: float
    leftSide: float
    rightSide: float


class RobotDefinition(TypedDict):
    """A python version of the robot definition type."""

    displayName: str
    robotType: RobotType
    models: List[str]
    extents: List[float]
    paddingOffsets: paddingOffset
    mountOffsets: mountOffset



from pathlib import Path
from typing import List


def list_v2_defs() -> List[str]:
    return [
        deffile.stem
        for deffile in (
            Path(__file__).parent / ".." / ".." / "module" / "definitions" / "2"
        ).iterdir()
    ]


def list_v3_defs() -> List[str]:
    return [
        deffile.stem
        for deffile in (
            Path(__file__).parent / ".." / ".." / "module" / "definitions" / "3"
        ).iterdir()
    ]



import pytest
import typeguard

from opentrons_shared_data.deck import (
    list_names as list_deck_definition_names,
    load as load_deck_definition,
)
from opentrons_shared_data.deck.types import (
    DeckDefinitionV3,
    DeckDefinitionV5,
)


@pytest.mark.parametrize("defname", list_deck_definition_names(version=3))
def test_v3_defs(defname: str) -> None:
    defn = load_deck_definition(name=defname, version=3)
    typeguard.check_type(defn, DeckDefinitionV3)


@pytest.mark.parametrize("defname", list_deck_definition_names(version=5))
def test_v5_defs(defname: str) -> None:
    defn = load_deck_definition(name=defname, version=5)
    typeguard.check_type(defn, DeckDefinitionV5)



"""
opentrons_shared_data.deck: types and bindings for deck definitions
"""
from typing import Dict, List, NamedTuple, cast, overload, TYPE_CHECKING
from typing_extensions import Final
import json

from .. import get_shared_data_root, load_shared_data

if TYPE_CHECKING:
    from .types import (
        DeckSchema,
        DeckDefinition,
        DeckDefinitionV3,
        DeckSchemaVersion3,
        DeckDefinitionV4,
        DeckSchemaVersion4,
        DeckDefinitionV5,
        DeckSchemaVersion5,
    )

DEFAULT_DECK_DEFINITION_VERSION: Final = 5


class Offset(NamedTuple):
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0


Z_PREP_OFFSET = Offset(x=13, y=13)
CALIBRATION_PROBE_RADIUS: Final[float] = 2
CALIBRATION_SQUARE_DEPTH: Final[float] = -0.25
CALIBRATION_SQUARE_SIZE: Final[float] = 20
CALIBRATION_SQUARE_EDGES: Dict[str, Offset] = {
    "left": Offset(x=-CALIBRATION_SQUARE_SIZE * 0.5),
    "right": Offset(x=CALIBRATION_SQUARE_SIZE * 0.5),
    "top": Offset(y=CALIBRATION_SQUARE_SIZE * 0.5),
    "bottom": Offset(y=-CALIBRATION_SQUARE_SIZE * 0.5),
}


@overload
def load(name: str, version: "DeckSchemaVersion5") -> "DeckDefinitionV5":
    ...


@overload
def load(name: str, version: "DeckSchemaVersion4") -> "DeckDefinitionV4":
    ...


@overload
def load(name: str, version: "DeckSchemaVersion3") -> "DeckDefinitionV3":
    ...


@overload
def load(name: str) -> "DeckDefinitionV5":
    ...


def load(name: str, version: int = DEFAULT_DECK_DEFINITION_VERSION) -> "DeckDefinition":
    return json.loads(  # type: ignore[no-any-return]
        load_shared_data(f"deck/definitions/{version}/{name}.json")
    )


def load_schema(version: int) -> "DeckSchema":
    return cast(
        "DeckSchema", json.loads(load_shared_data(f"deck/schemas/{version}.json"))
    )


def list_names(version: int) -> List[str]:
    """Return all loadable deck definition names, for the given schema version."""
    definitions_directory = (
        get_shared_data_root() / "deck" / "definitions" / f"{version}"
    )
    return [file.stem for file in definitions_directory.iterdir()]


def get_calibration_square_position_in_slot(slot: int) -> Offset:
    """Get the position of an OT-3 deck slot's calibration square.

    Params:
        slot: The slot whose calibration square to retrieve, specified as an OT-2-style slot number.
            For example, specify 5 to get slot C2.
    """
    deck = load("ot3_standard", version=3)
    slots = deck["locations"]["orderedSlots"]

    # Assume that the OT-3 deck definition has the same number of slots, and in the same order,
    # as the OT-2.
    # TODO(mm, 2023-05-22): This assumption will break down when the OT-3 has staging area slots.
    # https://opentrons.atlassian.net/browse/RLAB-345
    s = slots[slot - 1]

    bottom_left = s["position"]
    slot_size_x = s["boundingBox"]["xDimension"]
    slot_size_y = s["boundingBox"]["yDimension"]
    relative_center = [float(slot_size_x) * 0.5, float(slot_size_y) * 0.5, 0]
    square_z = [0, 0, CALIBRATION_SQUARE_DEPTH]
    # add up the elements of each list and return an Offset of the result
    nominal_position = [
        float(sum(x)) for x in zip(bottom_left, relative_center, square_z)
    ]
    return Offset(*nominal_position)



"""A Python package wrapping json config definitions for the Opentrons stack.

This package should never be installed on its own, only as a dependency of
the main opentrons package
"""

import os

from .load import get_shared_data_root, load_shared_data

from ._version import version

HERE = os.path.abspath(os.path.dirname(__file__))

__version__ = version

__all__ = ["__version__", "get_shared_data_root", "load_shared_data"]



"""Types and functions for accessing liquid class definitions."""
from pathlib import Path
from .. import load_shared_data, get_shared_data_root
from .liquid_class_definition import LiquidClassSchemaV1


DEFAULT_SCHEMA_VERSION = 1
DEFAULT_LC_VERSION = 1


class LiquidClassDefinitionDoesNotExist(Exception):
    """Specified liquid class definition does not exist."""


def load_definition(
    name: str,
    version: int = DEFAULT_LC_VERSION,
    schema_version: int = DEFAULT_SCHEMA_VERSION,
) -> LiquidClassSchemaV1:
    """Load the specified liquid class' definition as a LiquidClassSchemaV1 object.

    Note: this is an expensive operation and should be called sparingly.
    """
    try:
        return LiquidClassSchemaV1.model_validate_json(
            load_shared_data(
                f"liquid-class/definitions/{schema_version}/{name}/{version}.json"
            )
        )
    except FileNotFoundError:
        raise LiquidClassDefinitionDoesNotExist(
            f"No definition found for liquid class '{name}' version {version}"
        )


def definition_exists(
    name: str,
    version: int = DEFAULT_LC_VERSION,
) -> bool:
    """Return whether a definition exists for the specified liquid class name.."""
    return Path(
        get_shared_data_root()
        / f"liquid-class/definitions/{DEFAULT_SCHEMA_VERSION}/{name}/{version}.json"
    ).exists()



"""Tests that validate the built-in liquid class definitions."""
import pytest
from typing import List, Dict, Any

from opentrons_shared_data import get_shared_data_root
from opentrons_shared_data.liquid_classes import load_definition
from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    DelayProperties,
    MixProperties,
    TouchTipProperties,
    BlowoutProperties,
    BlowoutLocation,
    AspirateProperties,
    PositionReference,
    SingleDispenseProperties,
    MultiDispenseProperties,
    TransferProperties,
)


def _get_all_liquid_classes() -> List[str]:
    return [
        deffile.stem
        for deffile in (
            get_shared_data_root() / "liquid-class" / "definitions" / "1"
        ).iterdir()
    ]


@pytest.mark.parametrize("liquid_class_name", list(_get_all_liquid_classes()))
def test_validate_unique_pipette_keys(liquid_class_name: str) -> None:
    """A liquid class definition should contain only one set of properties per pipette model."""
    definition_dict = load_definition(liquid_class_name, version=1, schema_version=1)
    pipette_models = [prop.pipetteModel for prop in definition_dict.byPipette]
    assert len(pipette_models) == len(set(pipette_models))


@pytest.mark.parametrize("liquid_class_name", list(_get_all_liquid_classes()))
def test_validate_unique_tip_keys(liquid_class_name: str) -> None:
    """A liquid class definition should contain only one set of properties per tip type."""
    definition_dict = load_definition(liquid_class_name, version=1, schema_version=1)

    for by_pip_prop in definition_dict.byPipette:
        tipracks = [tip_prop.tiprack for tip_prop in by_pip_prop.byTipType]
        assert len(tipracks) == len(set(tipracks))


def test_validate_delay_properties_dict() -> None:
    """Delay properties model validator should convert valid dict to DelayProperties."""
    obj = DelayProperties.model_validate({"enable": True, "duration": 2})
    assert isinstance(obj, DelayProperties)
    assert obj.enable is True
    assert obj.params.duration == 2  # type: ignore[union-attr]

    with pytest.raises(
        ValueError,
        match="Delay properties should specify either duration or params, not both",
    ):
        DelayProperties.model_validate(
            {"enable": False, "params": {"duration": 2}, "duration": 2}
        )


def test_validate_mix_properties_dict() -> None:
    """Mix properties model validator should convert valid dict to MixProperties."""
    obj = MixProperties.model_validate({"enable": True, "repetitions": 2, "volume": 3})
    assert isinstance(obj, MixProperties)
    assert obj.enable is True
    assert obj.params is not None
    assert obj.params.repetitions == 2
    assert obj.params.volume == 3

    with pytest.raises(
        ValueError,
        match="either all of \\['repetitions', 'volume'\\] or 'params', not both",
    ):
        MixProperties.model_validate(
            {"enable": True, "repetitions": 2, "volume": 3, "params": {"foo": "bar"}}
        )


def test_validate_touchtip_properties_dict() -> None:
    """Touch tip properties model validator should convert valid dict to TouchTipProperties."""
    obj = TouchTipProperties.model_validate(
        {"enable": True, "z_offset": 2, "mm_from_edge": 3, "speed": 4}
    )
    assert isinstance(obj, TouchTipProperties)
    assert obj.enable is True
    assert obj.params is not None
    assert obj.params.mmFromEdge == 3
    assert obj.params.speed == 4
    assert obj.params.zOffset == 2

    with pytest.raises(
        ValueError,
        match="either all of \\['z_offset', 'mm_from_edge', 'speed'\\] or 'params', not both",
    ):
        TouchTipProperties.model_validate(
            {
                "enable": True,
                "z_offset": 2,
                "mm_from_edge": 3,
                "speed": 4,
                "params": {"foo": "bar"},
            }
        )


def test_validate_blowout_properties_dict() -> None:
    """Blowout properties model validator should convert valid dict to BlowoutProperties."""
    obj = BlowoutProperties.model_validate(
        {"enable": True, "location": "source", "flow_rate": 3}
    )
    assert isinstance(obj, BlowoutProperties)
    assert obj.enable is True
    assert obj.params is not None
    assert obj.params.location == BlowoutLocation.SOURCE
    assert obj.params.flowRate == 3

    with pytest.raises(
        ValueError,
        match="either all of \\['location', 'flow_rate'\\] or 'params', not both",
    ):
        BlowoutProperties.model_validate(
            {
                "enable": True,
                "location": "source",
                "flow_rate": 3,
                "params": {"foo": "bar"},
            }
        )


def test_validate_aspirate_properties_dict(
    sample_transfer_properties_dict: Dict[str, Dict[str, Any]],
) -> None:
    """Aspirate properties model validator should convert valid dict to AspirateProperties."""
    obj = AspirateProperties.model_validate(
        sample_transfer_properties_dict["flex_1channel_50"][
            "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ]["aspirate"]
    )
    assert isinstance(obj, AspirateProperties)
    assert obj.aspiratePosition.positionReference == PositionReference.WELL_BOTTOM
    assert obj.mix.enable is False


def test_validate_single_dispense_properties_dict(
    sample_transfer_properties_dict: Dict[str, Dict[str, Any]],
) -> None:
    """Single dispense properties model validator should convert valid dict to SingleDispenseProperties."""
    obj = SingleDispenseProperties.model_validate(
        sample_transfer_properties_dict["flex_1channel_50"][
            "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ]["dispense"]
    )
    assert isinstance(obj, SingleDispenseProperties)
    assert obj.dispensePosition.positionReference == PositionReference.WELL_BOTTOM
    assert obj.mix.enable is False


def test_validate_multi_dispense_properties_dict(
    sample_transfer_properties_dict: Dict[str, Dict[str, Any]],
) -> None:
    """Multi dispense properties model validator should convert valid dict to MultiDispenseProperties."""
    obj = MultiDispenseProperties.model_validate(
        sample_transfer_properties_dict["flex_1channel_50"][
            "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ]["multi_dispense"]
    )
    assert isinstance(obj, MultiDispenseProperties)
    assert obj.dispensePosition.positionReference == PositionReference.WELL_BOTTOM
    assert obj.conditioningByVolume == [(0, 0)]
    assert obj.disposalByVolume == [(0, 5)]


def test_validate_transfer_properties_dict(
    sample_transfer_properties_dict: Dict[str, Dict[str, Any]],
) -> None:
    """Transfer properties model validator should convert valid dict to TransferProperties."""
    obj = TransferProperties.model_validate(
        sample_transfer_properties_dict["flex_1channel_50"][
            "opentrons/opentrons_flex_96_tiprack_50ul/1"
        ]
    )
    assert isinstance(obj, TransferProperties)
    assert (
        obj.aspirate.aspiratePosition.positionReference == PositionReference.WELL_BOTTOM
    )
    assert obj.singleDispense.pushOutByVolume == [(10.0, 7.0), (20.0, 10.0)]
    assert obj.multiDispense.conditioningByVolume == [(0, 0)]  # type: ignore[union-attr]
    assert obj.multiDispense.disposalByVolume == [(0, 5)]  # type: ignore[union-attr]



"""Test exception handling."""
from opentrons_shared_data.errors.exceptions import PythonException


def test_exception_wrapping_single_level() -> None:
    """When we make a PythonException it should capture details of an exception."""
    try:
        raise RuntimeError("oh no!")
    except RuntimeError as e:
        captured_exc = e

    wrapped = PythonException(captured_exc)
    assert wrapped.detail["class"] == "RuntimeError"
    assert wrapped.detail["traceback"]
    assert wrapped.detail["args"] == "('oh no!',)"


def test_exception_wrapping_multi_level() -> None:
    """We can wrap all exceptions in a raise-from chain."""

    def _raise_inner() -> None:
        raise KeyError("oh no!")

    try:
        try:
            _raise_inner()
        except KeyError as e:
            raise RuntimeError("uh oh!") from e
    except RuntimeError as e:
        captured_exc = e

    wrapped = PythonException(captured_exc)
    assert wrapped.detail["class"] == "RuntimeError"
    assert wrapped.detail["traceback"]
    assert wrapped.detail["args"] == "('uh oh!',)"
    second = wrapped.wrapping[0]
    assert isinstance(second, PythonException)
    assert second.detail["class"] == "KeyError"
    assert second.detail["args"] == "('oh no!',)"
    assert not second.wrapping


def test_exception_wrapping_rethrow() -> None:
    """We can wrap raise-froms."""
    try:
        try:
            raise RuntimeError("oh no!")
        except RuntimeError as e:
            raise PythonException(e) from e
    except PythonException as pe:
        wrapped = pe

    assert wrapped.detail["class"] == "RuntimeError"
    assert wrapped.detail["traceback"]
    assert wrapped.detail["args"] == "('oh no!',)"


def test_exception_wrapping_error_in_except() -> None:
    """We can even wrap exceptions in exception handlers."""
    try:
        try:
            raise RuntimeError("oh no!")
        except RuntimeError:
            raise KeyError("uh oh")
    except BaseException as e:
        wrapped = PythonException(e)

    # when we have an exception-during-handler, the originally-handled
    # exception is what's bubbled up:
    assert wrapped.detail["class"] == "KeyError"
    assert len(wrapped.wrapping) == 1
    assert wrapped.wrapping[0].detail["class"] == "RuntimeError"
    assert wrapped.wrapping[0].detail["traceback"]
    assert wrapped.wrapping[0].detail["args"] == "('oh no!',)"



"""Tests for the default values in built-in liquid class definitions."""
import pytest
from typing import List

from opentrons_shared_data import get_shared_data_root
from opentrons_shared_data.liquid_classes import load_definition


def _get_all_liquid_classes() -> List[str]:
    return [
        deffile.stem
        for deffile in (
            get_shared_data_root() / "liquid-class" / "definitions" / "1"
        ).iterdir()
    ]


@pytest.mark.parametrize("liquid_class_name", list(_get_all_liquid_classes()))
def test_correction_values_equal_each_other(liquid_class_name: str) -> None:
    """The correction volume values for each pipette/tiprack combo should all be equal to one another."""
    liquid_class_def = load_definition(liquid_class_name)
    incorrect_combos = []
    for by_pipette in liquid_class_def.byPipette:
        for liquid_class_props in by_pipette.byTipType:
            if liquid_class_props.multiDispense is None:
                continue
            aspirate_correction = liquid_class_props.aspirate.correctionByVolume
            dispense_correction = liquid_class_props.singleDispense.correctionByVolume
            multi_dispense_correction = (
                liquid_class_props.multiDispense.correctionByVolume
            )
            if (
                aspirate_correction != dispense_correction
                or dispense_correction != multi_dispense_correction
            ):
                incorrect_combos.append(
                    f"{by_pipette.pipetteModel} {liquid_class_props.tiprack}"
                )
    assert incorrect_combos == []


@pytest.mark.parametrize("liquid_class_name", list(_get_all_liquid_classes()))
def test_correction_volume_not_negative(liquid_class_name: str) -> None:
    """Correction volumes must not push the plunger below zero position or else PipetteNotReadyToAspirateError."""
    liquid_class_def = load_definition(liquid_class_name)
    for by_pipette in liquid_class_def.byPipette:
        for by_tip_type in by_pipette.byTipType:
            # We just need to check one of aspirate/singleDispense/multiDispense, because
            # the test above ensures that all 3 are the same.
            correction_by_volume = by_tip_type.aspirate.correctionByVolume
            # The correction volume at 0 must be 0:
            assert (
                0.0,
                0.0,
            ) in correction_by_volume, f"Correction volume not 0 at 0 in {by_pipette.pipetteModel} {by_tip_type.tiprack}"
            # The nominal volume + correction volume must never be below 0.
            # (Seth thinks this check is sufficient to ensure that the plunger won't
            # go below the zero position, whereas David thinks we need a stronger
            # check that takes into account ul_per_mm(). But this is a start.)
            for nominal_volume, correction_volume in correction_by_volume:
                assert (
                    nominal_volume + correction_volume >= 0
                ), f"Volume + correction volume is negative in {by_pipette.pipetteModel} {by_tip_type.tiprack}"


@pytest.mark.parametrize("liquid_class_name", list(_get_all_liquid_classes()))
def test_flow_rates_equal_each_other(liquid_class_name: str) -> None:
    """The dispense flow rate values for each pipette/tiprack combo should all be equal to one another."""
    liquid_class_def = load_definition(liquid_class_name)
    incorrect_combos = []
    for by_pipette in liquid_class_def.byPipette:
        for liquid_class_props in by_pipette.byTipType:
            if liquid_class_props.multiDispense is None:
                continue
            dispense_flow_rates = liquid_class_props.singleDispense.flowRateByVolume
            multi_dispense_flow_rates = (
                liquid_class_props.multiDispense.flowRateByVolume
            )
            if dispense_flow_rates != multi_dispense_flow_rates:
                incorrect_combos.append(
                    f"{by_pipette.pipetteModel} {liquid_class_props.tiprack}"
                )
    assert incorrect_combos == []



from pydantic import ConfigDict, BaseModel, Field
from typing import Any, List, Optional, Dict, Union
from typing_extensions import Literal

from opentrons_shared_data.labware.labware_definition import LabwareDefinition2

from .shared_models import (
    Liquid,
    CommandAnnotation,
    Location,
    ProfileStep,
    WellLocation,
    OffsetVector,
    Metadata,
    Robot,
    DesignerApplication,
)


# TODO (tamar 3/15/22): split apart all the command payloads when we tackle #9583
class Params(BaseModel):
    slotName: Optional[str] = None
    axes: Optional[List[str]] = None
    pipetteId: Optional[str] = None
    mount: Optional[str] = None
    moduleId: Optional[str] = None
    location: Optional[Union[Location, Literal["offDeck"]]] = None
    labwareId: Optional[str] = None
    displayName: Optional[str] = None
    liquidId: Optional[str] = None
    volumeByWell: Optional[Dict[str, Any]] = None
    wellName: Optional[str] = None
    volume: Optional[float] = None
    flowRate: Optional[float] = None
    wellLocation: Optional[WellLocation] = None
    waitForResume: Optional[Literal[True]] = None
    seconds: Optional[float] = None
    minimumZHeight: Optional[float] = None
    forceDirect: Optional[bool] = None
    speed: Optional[float] = None
    message: Optional[str] = None
    coordinates: Optional[OffsetVector] = None
    axis: Optional[str] = None
    distance: Optional[float] = None
    positionId: Optional[str] = None
    temperature: Optional[float] = None
    celsius: Optional[float] = None
    blockMaxVolumeUl: Optional[float] = None
    rpm: Optional[float] = None
    height: Optional[float] = None
    offset: Optional[OffsetVector] = None
    profile: Optional[List[ProfileStep]] = None
    radius: Optional[float] = None
    # schema v7 add-ons
    newLocation: Optional[Union[Location, Literal["offDeck"]]] = None
    strategy: Optional[str] = None
    pickUpOffset: Optional[OffsetVector] = None
    dropOffset: Optional[OffsetVector] = None
    homeAfter: Optional[bool] = None
    alternateDropLocation: Optional[bool] = None
    holdTimeSeconds: Optional[float] = None
    maintenancePosition: Optional[str] = None
    pipetteName: Optional[str] = None
    model: Optional[str] = None
    loadName: Optional[str] = None
    namespace: Optional[str] = None
    version: Optional[int] = None
    pushOut: Optional[float] = None


class Command(BaseModel):
    commandType: str
    params: Params
    key: Optional[str] = None


class ProtocolSchemaV7(BaseModel):
    otSharedSchema: Literal["#/protocol/schemas/7"] = Field(
        ...,
        alias="$otSharedSchema",
        description="The path to a valid Opentrons shared schema relative to "
        "the shared-data directory, without its extension.",
    )
    schemaVersion: Literal[7]
    metadata: Metadata
    robot: Robot
    liquids: Optional[Dict[str, Liquid]] = None
    labwareDefinitions: Dict[str, LabwareDefinition2]
    commands: List[Command]
    commandAnnotations: Optional[List[CommandAnnotation]] = None
    designerApplication: Optional[DesignerApplication] = None
    model_config = ConfigDict(populate_by_name=True)



"""
opentrons_shared_data.protocol: types and bindings for json protocols
"""
import json
from typing import Any, NewType, Dict

from .. import load_shared_data

Schema = NewType("Schema", Dict[str, Any])


def load_schema(version: int) -> "Schema":
    return json.loads(load_shared_data(f"protocol/schemas/{version}.json"))
